# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for API tests."""
from keystonemiddleware import fixture
from oslo_utils import uuidutils
import pecan
import pecan.testing

from blazar.api import context as api_context
from blazar import context
from blazar.manager.leases import rpcapi as leases_api
from blazar.manager.oshosts import rpcapi as hosts_rpcapi
from blazar import tests

PATH_PREFIX = '/v2'

FAKE_PROJECT = '981b767265174c108bc5a61185b748ac'
FAKE_USER = 'b4fdb2fff13545ceb751295096cc18ee'


class APITest(tests.TestCase):
    """Used for unittests tests of Pecan controllers."""

    ADMIN_TOKEN = uuidutils.generate_uuid(dashed=False)
    MEMBER_TOKEN = uuidutils.generate_uuid(dashed=False)

    # SOURCE_DATA = {'test_source': {'somekey': '666'}}

    def setUp(self):
        def fake_ctx_from_headers(headers):
            if not headers:
                return context.BlazarContext(
                    user_id=FAKE_USER, project_id=FAKE_PROJECT,
                    roles=['member'])
            roles = headers.get('X-Roles', str('member')).split(',')
            return context.BlazarContext(
                user_id=headers.get('X-User-Id', FAKE_USER),
                project_id=headers.get('X-Project-Id', FAKE_PROJECT),
                auth_token=headers.get('X-Auth-Token', None),
                service_catalog=None,
                user_name=headers.get('X-User-Name', 'fake'),
                project_name=headers.get('X-Project-Name', 'fake'),
                roles=roles,
            )

        super(APITest, self).setUp()

        self.keystone_middleware = None
        self.app = self._make_app()

        # NOTE(sbauza): Context is taken from Keystone auth middleware, we need
        #               to simulate here
        self.api_context = api_context
        self.fake_ctx_from_headers = self.patch(self.api_context,
                                                'ctx_from_headers')
        self.fake_ctx_from_headers.side_effect = fake_ctx_from_headers

        self.rpcapi = leases_api.ManagerRPCAPI
        self.hosts_rpcapi = hosts_rpcapi.ManagerRPCAPI

        # self.patch(rpcapi.ManagerRPCAPI, 'list_leases').return_value = []

        def reset_pecan():
            pecan.set_config({}, overwrite=True)

        self.addCleanup(reset_pecan)

    def _make_app(self, enable_acl=False):
        # Determine where we are so we can set up paths in the config

        # NOTE(sbauza): Keystone middleware auth can be deactivated using
        #               enable_acl set to False

        if enable_acl:
            self.keystone_middleware = self.useFixture(
                fixture.AuthTokenFixture())

            self.keystone_middleware.add_token_data(token_id=self.ADMIN_TOKEN,
                                                    user_id='user_id1',
                                                    user_name='user_name1',
                                                    project_id='123i2910',
                                                    role_list=['admin'])
            self.keystone_middleware.add_token_data(token_id=self.MEMBER_TOKEN,
                                                    user_id='user_id2',
                                                    user_name='user-good',
                                                    project_id='project-good',
                                                    role_list=['Member'])

        self.config = {
            'app': {
                'modules': ['blazar.api.v2'],
                'root': 'blazar.api.root.RootController',
                'enable_acl': enable_acl,
            },
        }

        return pecan.testing.load_test_app(self.config)

    def _request_json(self, path, params, expect_errors=False, headers=None,
                      method="post", extra_environ=None, status=None,
                      path_prefix=PATH_PREFIX):
        """Sends simulated HTTP request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param method: Request method type. Appropriate method function call
                       should be used rather than passing attribute in.
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        :param path_prefix: prefix of the url path
        """
        full_path = path_prefix + path
        print('%s: %s %s' % (method.upper(), full_path, params))
        response = getattr(self.app, "%s_json" % method)(
            str(full_path),
            params=params,
            headers=headers,
            status=status,
            extra_environ=extra_environ,
            expect_errors=expect_errors
        )
        print('GOT:%s' % response)
        return response

    def put_json(self, path, params, expect_errors=False, headers=None,
                 extra_environ=None, status=None):
        """Sends simulated HTTP PUT request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="put")

    def post_json(self, path, params, expect_errors=False, headers=None,
                  extra_environ=None, status=None):
        """Sends simulated HTTP POST request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="post")

    def delete(self, path, expect_errors=False, headers=None,
               extra_environ=None, status=None, path_prefix=PATH_PREFIX):
        """Sends simulated HTTP DELETE request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        :param path_prefix: prefix of the url path
        """
        full_path = path_prefix + path
        print('DELETE: %s' % (full_path))
        response = self.app.delete(str(full_path),
                                   headers=headers,
                                   status=status,
                                   extra_environ=extra_environ,
                                   expect_errors=expect_errors)
        print('GOT:%s' % response)
        return response

    def get_json(self, path, expect_errors=False, headers=None,
                 extra_environ=None, q=[], path_prefix=PATH_PREFIX, **params):
        """Sends simulated HTTP GET request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value;whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param q: list of queries consisting of: field, value, op, and type
                  keys
        :param path_prefix: prefix of the url path
        :param params: content for wsgi.input of request
        """
        full_path = path_prefix + path
        query_params = {'q.field': [],
                        'q.value': [],
                        'q.op': [],
                        }
        for query in q:
            for name in ['field', 'op', 'value']:
                query_params['q.%s' % name].append(query.get(name, ''))
        all_params = {}
        all_params.update(params)
        if q:
            all_params.update(query_params)
        print('GET: %s %r' % (full_path, all_params))
        response = self.app.get(full_path,
                                params=all_params,
                                headers=headers,
                                extra_environ=extra_environ,
                                expect_errors=expect_errors)
        if not expect_errors:
            response = response.json
        print('GOT:%s' % response)
        return response



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1.access import create as create_access_info
from keystoneauth1.identity import access
from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient import client as keystone_client
import netaddr
from oslo_config import cfg

from blazar import context
from blazar.manager import exceptions


CONF = cfg.CONF


def get_os_auth_host(conf):
    """Description

    Returns os_auth_host from conf, surrounded by brackets if IPv6.
    """
    os_auth_host = conf.os_auth_host
    if netaddr.valid_ipv6(os_auth_host, netaddr.core.INET_PTON):
        os_auth_host = "[%s]" % os_auth_host
    return os_auth_host


def client_kwargs(**_kwargs):
    kwargs = _kwargs.copy()

    ctx = kwargs.pop('ctx', None)
    username = kwargs.pop('username',
                          CONF.os_admin_username)
    password = kwargs.pop('password',
                          CONF.os_admin_password)
    project_name = kwargs.pop('project_name',
                              CONF.os_admin_project_name)
    user_domain_name = kwargs.pop('user_domain_name',
                                  CONF.os_admin_user_domain_name)
    project_domain_name = kwargs.pop('project_domain_name',
                                     CONF.os_admin_project_domain_name)
    trust_id = kwargs.pop('trust_id', None)
    auth_url = kwargs.pop('auth_url', None)
    region_name = kwargs.pop('region_name', CONF.os_region_name)
    if ctx is None:
        try:
            ctx = context.current()
        except RuntimeError:
            pass
    if ctx is not None:
        kwargs.setdefault('global_request_id', ctx.global_request_id)

    if auth_url is None:
        auth_url = "%s://%s:%s/%s/%s" % (CONF.os_auth_protocol,
                                         get_os_auth_host(CONF),
                                         CONF.os_auth_port,
                                         CONF.os_auth_prefix,
                                         CONF.os_auth_version)

    auth_kwargs = dict(
        auth_url=auth_url,
        username=username,
        password=password,
        user_domain_name=user_domain_name,
        project_domain_name=project_domain_name
    )

    if trust_id is not None:
        auth_kwargs.update(trust_id=trust_id)
    else:
        auth_kwargs.update(project_name=project_name)

    auth = v3.Password(**auth_kwargs)

    sess_kwargs = dict(
        auth=auth
    )

    if CONF.cafile:
        sess_kwargs.update(verify=CONF.cafile)

    sess = session.Session(**sess_kwargs)

    kwargs.setdefault('session', sess)
    kwargs.setdefault('region_name', region_name)
    return kwargs


def client_user_kwargs(**_kwargs):
    kwargs = _kwargs.copy()

    auth_url = kwargs.pop('auth_url', None)
    region_name = kwargs.pop('region_name', CONF.os_region_name)

    if auth_url is None:
        auth_url = "%s://%s:%s/%s/%s" % (CONF.os_auth_protocol,
                                         get_os_auth_host(CONF),
                                         CONF.os_auth_port,
                                         CONF.os_auth_prefix,
                                         CONF.os_auth_version)

    # Pass the auth token present on the context directly on to the
    # next service; this effectively proxies the user's token they used
    # to authenticate to Blazar, and prevents having to re-authenticate
    # (which has issues for certain auth types, such as application creds)
    ctx = context.current()
    admin_ks_client = keystone_client.Client(
        version='3',
        **client_kwargs(**_kwargs)
    )
    data = admin_ks_client.tokens.get_token_data(ctx.auth_token)
    access_info = create_access_info(body=data, auth_token=ctx.auth_token)
    auth = access.AccessInfoPlugin(access_info, auth_url=auth_url)

    sess_kwargs = dict(
        auth=auth
    )

    if CONF.cafile:
        sess_kwargs.update(verify=CONF.cafile)

    sess = session.Session(**sess_kwargs)

    kwargs.setdefault('session', sess)
    kwargs.setdefault('region_name', region_name)
    return kwargs


def url_for(service_catalog, service_type, admin=False,
            endpoint_interface=None,
            os_region_name=None):
    """Description

    Gets url of the service to communicate through.
    service_catalog - dict contains info about specific OpenStack service
    service_type - OpenStack service type specification
    """
    if not endpoint_interface:
        if service_type == 'identity':
            endpoint_interface = CONF.endpoint_type
        elif service_type == 'compute':
            endpoint_interface = CONF.nova.endpoint_type
        else:
            endpoint_interface = 'public'

    if not isinstance(service_catalog, list):
        service_catalog = service_catalog.normalize_catalog()

    service = None
    for srv in service_catalog:
        if srv['type'] == service_type:
            service = srv

    if service:
        try:
            endpoints = service['endpoints']
        except KeyError:
            raise exceptions.EndpointsNotFound(
                "No endpoints for %s" % service['type'])
        if os_region_name:
            endpoints = [e for e in endpoints if e['region'] == os_region_name]
            if not endpoints:
                raise exceptions.EndpointsNotFound("No endpoints for %s in "
                                                   "region %s" %
                                                   (service['type'],
                                                    os_region_name))
        try:
            # if Keystone API v3 endpoints returned
            endpoint = [e for e in endpoints
                        if e['interface'] == endpoint_interface][0]
            return endpoint['url']
        except KeyError:
            # otherwise
            return endpoints[0]['%sURL' % endpoint_interface]
    else:
        raise exceptions.ServiceNotFound(
            'Service "%s" not found' % service_type)



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test of Policy Engine For Blazar."""

from oslo_config import cfg

from blazar import context
from blazar import exceptions
from blazar import policy
from blazar import tests

CONF = cfg.CONF


class BlazarPolicyTestCase(tests.TestCase):

    def setUp(self):
        super(BlazarPolicyTestCase, self).setUp()

        self.context = context.BlazarContext(user_id='fake',
                                             project_id='fake',
                                             roles=['member'])

    def test_standardpolicy(self):
        target_good = {'user_id': self.context.user_id,
                       'project_id': self.context.project_id}
        target_wrong = {'user_id': self.context.user_id,
                        'project_id': 'bad_project'}
        action = "blazar:leases:get"
        self.assertTrue(policy.enforce(self.context, action,
                                       target_good))
        self.assertFalse(policy.enforce(self.context, action,
                                        target_wrong, False))

    def test_adminpolicy(self):
        target = {'user_id': self.context.user_id,
                  'project_id': self.context.project_id}
        action = "blazar:oshosts:get"
        self.assertRaises(exceptions.PolicyNotAuthorized, policy.enforce,
                          self.context, action, target)

    def test_authorize(self):

        @policy.authorize('leases', 'get', ctx=self.context)
        def user_method_with_action(self):
            return True

        @policy.authorize('oshosts', 'get', ctx=self.context)
        def adminonly_method_with_action(self):
            return True

        self.assertTrue(user_method_with_action(self))
        self.assertRaises(exceptions.PolicyNotAuthorized,
                          adminonly_method_with_action, self)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import blazar.api.v2.app
import blazar.api.v2.controllers
import blazar.cmd.api
import blazar.config
import blazar.db.base
import blazar.db.migration.cli
import blazar.manager
import blazar.manager.service
import blazar.notification.notifier
import blazar.plugins.oshosts.host_plugin
import blazar.utils.openstack.keystone
import blazar.utils.openstack.nova


def list_opts():
    return [
        ('DEFAULT',
         itertools.chain(
             blazar.api.v2.app.auth_opts,
             blazar.cmd.api.opts,
             blazar.config.api_opts,
             blazar.config.cli_opts,
             blazar.config.lease_opts,
             blazar.config.os_opts,
             blazar.db.base.db_driver_opts,
             blazar.db.migration.cli.command_opts,
             blazar.utils.openstack.keystone.opts,
             blazar.utils.openstack.keystone.keystone_opts)),
        ('api', blazar.api.v2.controllers.api_opts),
        ('manager', itertools.chain(blazar.manager.opts,
                                    blazar.manager.service.manager_opts)),
        ('enforcement', itertools.chain(
            blazar.enforcement.filters.external_service_filter
            .ExternalServiceFilter.enforcement_opts,
            blazar.enforcement.filters.max_lease_duration_filter.MaxLeaseDurationFilter.enforcement_opts, # noqa
            blazar.enforcement.enforcement.enforcement_opts)),
        ('notifications', blazar.notification.notifier.notification_opts),
        ('nova', blazar.utils.openstack.nova.nova_opts),
        (blazar.plugins.oshosts.RESOURCE_TYPE,
         blazar.plugins.oshosts.host_plugin.plugin_opts),
    ]



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from blazar.api.v1.leases import service
from blazar.api.v1 import utils as api_utils
from blazar.api.v1 import validation
from blazar import utils

LOG = logging.getLogger(__name__)


def get_rest():
    """Return Rest app"""
    return rest


rest = api_utils.Rest('v1_0', __name__, url_prefix='/v1')
_api = utils.LazyProxy(service.API)


# Leases operations

@rest.get('/leases', query=True)
def leases_list(req, query):
    """List all existing leases."""
    return api_utils.render(leases=_api.get_leases(query))


@rest.post('/leases')
def leases_create(req, data):
    """Create new lease."""
    return api_utils.render(lease=_api.create_lease(data))


@rest.get('/leases/<lease_id>')
@validation.check_exists(_api.get_lease, lease_id='lease_id')
def leases_get(req, lease_id):
    """Get lease by its ID."""
    return api_utils.render(lease=_api.get_lease(lease_id))


@rest.put('/leases/<lease_id>')
@validation.check_exists(_api.get_lease, lease_id='lease_id')
def leases_update(req, lease_id, data):
    """Update lease."""
    return api_utils.render(lease=_api.update_lease(lease_id, data))


@rest.delete('/leases/<lease_id>')
@validation.check_exists(_api.get_lease, lease_id='lease_id')
def leases_delete(req, lease_id):
    """Delete specified lease."""
    _api.delete_lease(lease_id)
    return api_utils.render()


# Plugins operations

@rest.get('/plugins')
def plugins_list():
    """List all possible plugins."""
    return api_utils.render(plugins=_api.get_plugins())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from blazar import tests
from blazar.tests import fake_requests
from blazar.utils.openstack import exceptions
from blazar.utils.openstack import placement

from oslo_config import cfg
from oslo_config import fixture as conf_fixture
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

CONF = cfg.CONF
PLACEMENT_MICROVERSION = 1.29


class TestPlacementClient(tests.TestCase):
    def setUp(self):
        super(TestPlacementClient, self).setUp()
        self.cfg = self.useFixture(conf_fixture.Config(CONF))
        self.cfg.config(os_auth_host='foofoo')
        self.cfg.config(os_auth_port='8080')
        self.cfg.config(os_auth_prefix='identity')
        self.cfg.config(os_auth_version='v3')
        self.cfg.config(os_region_name='region_foo')
        self.client = placement.BlazarPlacementClient()

    def test_client_auth_url(self):
        client = self.client._create_client()
        self.assertEqual("http://foofoo:8080/identity/v3",
                         client.session.auth.auth_url)

    def _add_default_kwargs(self, kwargs):
        kwargs['endpoint_filter'] = {'service_type': 'placement',
                                     'interface': 'internal',
                                     'region_name': 'region_foo'}
        kwargs['headers'] = {'accept': 'application/json'}
        kwargs['microversion'] = PLACEMENT_MICROVERSION
        kwargs['raise_exc'] = False
        kwargs['rate_semaphore'] = mock.ANY

    def _assert_keystone_called_once(self, kss_req, url, method, **kwargs):
        self._add_default_kwargs(kwargs)
        kss_req.assert_called_once_with(url, method, **kwargs)

    def _assert_keystone_called_any(self, kss_req, url, method, **kwargs):
        self._add_default_kwargs(kwargs)
        kss_req.assert_any_call(url, method, **kwargs)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_get(self, kss_req):
        kss_req.return_value = fake_requests.FakeResponse(200)
        url = '/resource_providers'
        resp = self.client.get(url)
        self.assertEqual(200, resp.status_code)
        self._assert_keystone_called_once(kss_req, url, 'GET')

    @mock.patch('keystoneauth1.session.Session.request')
    def test_post(self, kss_req):
        kss_req.return_value = fake_requests.FakeResponse(200)
        url = '/resource_providers'
        data = {'name': 'unicorn'}
        resp = self.client.post(url, data)
        self.assertEqual(200, resp.status_code)
        self._assert_keystone_called_once(kss_req, url, 'POST', json=data)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_put(self, kss_req):
        kss_req.return_value = fake_requests.FakeResponse(200)
        url = '/resource_providers'
        data = {'name': 'unicorn'}
        resp = self.client.put(url, data)
        self.assertEqual(200, resp.status_code)
        self._assert_keystone_called_once(kss_req, url, 'PUT', json=data)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete(self, kss_req):
        kss_req.return_value = fake_requests.FakeResponse(200)
        url = '/resource_providers'
        resp = self.client.delete(url)
        self.assertEqual(200, resp.status_code)
        self._assert_keystone_called_once(kss_req, url, 'DELETE')

    @mock.patch('keystoneauth1.session.Session.request')
    def test_get_resource_provider(self, kss_req):
        rp_name = 'blazar'
        rp_uuid = uuidutils.generate_uuid()
        parent_uuid = uuidutils.generate_uuid()

        mock_json_data = {
            'resource_providers': [
                {
                    'uuid': rp_uuid,
                    'name': rp_name,
                    'generation': 0,
                    'parent_provider_uuid': parent_uuid
                }
            ]
        }

        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_json_data))

        result = self.client.get_resource_provider(rp_name)

        expected_url = '/resource_providers?name=blazar'
        self._assert_keystone_called_once(kss_req, expected_url, 'GET')
        expected = {'uuid': rp_uuid,
                    'name': rp_name,
                    'generation': 0,
                    'parent_provider_uuid': parent_uuid}
        self.assertEqual(expected, result)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_get_resource_provider_no_rp(self, kss_req):
        rp_name = 'blazar'

        mock_json_data = {
            'resource_providers': []
        }

        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_json_data))

        result = self.client.get_resource_provider(rp_name)

        expected_url = '/resource_providers?name=blazar'
        self._assert_keystone_called_once(kss_req, expected_url, 'GET')
        self.assertEqual(None, result)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_get_resource_provider_fail(self, kss_req):
        rp_name = 'blazar'
        kss_req.return_value = fake_requests.FakeResponse(404)

        self.assertRaises(
            exceptions.ResourceProviderRetrievalFailed,
            self.client.get_resource_provider, rp_name)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_get_traits(self, kss_req):
        rp_uuid = uuidutils.generate_uuid()

        mock_json_data = {
            "resource_provider_generation": 1,
            "traits": [
                "CUSTOM_HW_FPGA_CLASS1",
                "CUSTOM_HW_FPGA_CLASS3"
            ]
        }

        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_json_data))

        result = self.client.get_traits(rp_uuid)

        expected_url = '/resource_providers/%s/traits' % rp_uuid
        self._assert_keystone_called_once(kss_req, expected_url, 'GET')
        expected = [
            "CUSTOM_HW_FPGA_CLASS1",
            "CUSTOM_HW_FPGA_CLASS3"
        ]
        self.assertEqual(expected, result)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_resource_provider(self, kss_req):
        rp_name = 'Blazar'
        rp_uuid = uuidutils.generate_uuid()
        parent_uuid = uuidutils.generate_uuid()

        mock_json_data = {'uuid': rp_uuid,
                          'name': rp_name,
                          'generation': 0,
                          'parent_provider_uuid': parent_uuid}

        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_json_data))

        result = self.client.create_resource_provider(
            rp_name, rp_uuid=rp_uuid, parent_uuid=parent_uuid)

        expected_url = '/resource_providers'
        expected_data = {'uuid': rp_uuid,
                         'name': rp_name,
                         'parent_provider_uuid': parent_uuid}
        self._assert_keystone_called_once(kss_req, expected_url, 'POST',
                                          json=expected_data)
        self.assertEqual(mock_json_data, result)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_resource_provider_fail(self, kss_req):
        rp_name = 'Blazar'
        kss_req.return_value = fake_requests.FakeResponse(404)

        self.assertRaises(
            exceptions.ResourceProviderCreationFailed,
            self.client.create_resource_provider, rp_name)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_resource_provider(self, kss_req):
        rp_uuid = uuidutils.generate_uuid()
        kss_req.return_value = fake_requests.FakeResponse(200)

        self.client.delete_resource_provider(rp_uuid)

        expected_url = '/resource_providers/' + str(rp_uuid)
        self._assert_keystone_called_once(kss_req, expected_url, 'DELETE')

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_resource_provider_fail(self, kss_req):
        rp_uuid = uuidutils.generate_uuid()
        kss_req.return_value = fake_requests.FakeResponse(404)

        self.assertRaises(
            exceptions.ResourceProviderDeletionFailed,
            self.client.delete_resource_provider, rp_uuid)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_reservation_provider(self, kss_req):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"
        get_json_mock = {
            'resource_providers': [
                {
                    'uuid': host_uuid,
                    'name': host_name,
                    'generation': 0,
                    'parent_provider_uuid': None
                }
            ]
        }
        post_json_mock = {'uuid': rp_uuid,
                          'name': rp_name,
                          'generation': 0,
                          'parent_provider_uuid': host_uuid}
        mock_call1 = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(get_json_mock))
        mock_call2 = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(post_json_mock))
        kss_req.side_effect = [mock_call1, mock_call2]

        self.client.create_reservation_provider(host_name)

        expected_url_get = "/resource_providers?name=%s" % host_name
        self._assert_keystone_called_any(kss_req, expected_url_get, 'GET')

        expected_url_post = "/resource_providers"
        expected_data = {'name': 'blazar_compute-1',
                         'parent_provider_uuid': host_uuid}
        self._assert_keystone_called_any(kss_req, expected_url_post, 'POST',
                                         json=expected_data)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_reservation_provider_fail(self, kss_req):
        host_name = "compute-1"
        get_json_mock = {'resource_providers': []}
        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(get_json_mock))
        self.assertRaises(
            exceptions.ResourceProviderNotFound,
            self.client.create_reservation_provider, host_name)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_provider(self, kss_req):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"
        get_json_mock = {
            'resource_providers': [
                {
                    'uuid': rp_uuid,
                    'name': rp_name,
                    'generation': 0,
                    'parent_provider_uuid': host_uuid
                }
            ]
        }
        mock_call1 = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(get_json_mock))
        mock_call2 = fake_requests.FakeResponse(200)
        kss_req.side_effect = [mock_call1, mock_call2]

        self.client.delete_reservation_provider(host_name)

        expected_url_get = "/resource_providers?name=%s" % rp_name
        self._assert_keystone_called_any(kss_req, expected_url_get, 'GET')

        expected_url_post = "/resource_providers/%s" % rp_uuid
        self._assert_keystone_called_any(kss_req, expected_url_post, 'DELETE')

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_provider_no_rp(self, kss_req):
        host_name = "compute-1"
        rp_name = "blazar_compute-1"
        get_json_mock = {
            'resource_providers': []
        }
        mock_call1 = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(get_json_mock))
        mock_call2 = fake_requests.FakeResponse(200)
        kss_req.side_effect = [mock_call1, mock_call2]

        self.client.delete_reservation_provider(host_name)

        expected_url_get = "/resource_providers?name=%s" % rp_name
        self._assert_keystone_called_any(kss_req, expected_url_get, 'GET')

        # Ensure that mock_call2 for delete is not called
        self.assertEqual(kss_req.call_count, 1)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_reservation_class(self, kss_req):
        rc_name = 'abc-def'
        kss_req.return_value = fake_requests.FakeResponse(200)

        self.client.create_reservation_class(rc_name)

        expected_url = '/resource_classes'
        expected_data = {'name': 'CUSTOM_RESERVATION_ABC_DEF'}
        self._assert_keystone_called_once(kss_req, expected_url, 'POST',
                                          json=expected_data)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_create_reservation_class_fail(self, kss_req):
        rc_name = 'abc-def'
        kss_req.return_value = fake_requests.FakeResponse(400)

        self.assertRaises(
            exceptions.ResourceClassCreationFailed,
            self.client.create_reservation_class, rc_name)

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_class(self, kss_req):
        rc_name = 'abc-def'
        kss_req.return_value = fake_requests.FakeResponse(200)

        self.client.delete_reservation_class(rc_name)

        expected_url = '/resource_classes/CUSTOM_RESERVATION_ABC_DEF'
        self._assert_keystone_called_once(kss_req, expected_url, 'DELETE')

    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_class_fail(self, kss_req):
        rc_name = 'abc-def'
        # If no reservation class found, the placement API returns 404 error.
        kss_req.return_value = fake_requests.FakeResponse(404)
        # Ensure that no error is raised
        self.client.delete_reservation_class(rc_name)

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get')
    @mock.patch('keystoneauth1.session.Session.request')
    def test_update_reservation_inventory(self, kss_req, client_get, get_rp):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"

        # Build the mock of current resource provider
        mock_get_rp_json = {'uuid': rp_uuid,
                            'name': rp_name,
                            'generation': 0,
                            'parent_provider_uuid': host_uuid}
        get_rp.return_value = mock_get_rp_json

        # Build the mock of "current" inventory for get_inventory()
        curr_gen = 11
        mock_get_inv_json = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 1
                },
            },
            "resource_provider_generation": curr_gen
        }
        client_get.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_get_inv_json))

        # Build the mock of "updated" inventory for update_inventory()
        update_gen = 12
        mock_put_json = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 1
                },
                'CUSTOM_RESERVATION_ADD': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                },
            },
            "resource_provider_generation": update_gen
        }
        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_put_json))

        result = self.client.update_reservation_inventory(host_name, 'add', 3)

        expected_data = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 1
                },
                'CUSTOM_RESERVATION_ADD': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                },
            },
            "resource_provider_generation": curr_gen
        }
        expected_url = '/resource_providers/%s/inventories' % rp_uuid
        self._assert_keystone_called_once(kss_req, expected_url, 'PUT',
                                          json=expected_data)
        self.assertEqual(mock_put_json, result)

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get')
    @mock.patch('keystoneauth1.session.Session.request')
    def test_add_reservation_inventory(self, kss_req, client_get, get_rp):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"

        # Build the mock of current resource provider
        mock_get_rp_json = {'uuid': rp_uuid,
                            'name': rp_name,
                            'generation': 0,
                            'parent_provider_uuid': host_uuid}
        get_rp.return_value = mock_get_rp_json

        # Build the mock of "current" inventory for get_inventory()
        curr_gen = 11
        mock_get_inv_json = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 1
                },
            },
            "resource_provider_generation": curr_gen
        }
        client_get.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_get_inv_json))

        # Build the mock of "updated" inventory for update_inventory()
        update_gen = 12
        mock_put_json = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                },
            },
            "resource_provider_generation": update_gen
        }
        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_put_json))

        result = self.client.update_reservation_inventory(
            host_name, 'curr', 2, additional=True)

        expected_data = {
            'inventories': {
                'CUSTOM_RESERVATION_CURR': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                }
            },
            "resource_provider_generation": curr_gen
        }
        expected_url = '/resource_providers/%s/inventories' % rp_uuid
        self._assert_keystone_called_once(kss_req, expected_url, 'PUT',
                                          json=expected_data)
        self.assertEqual(mock_put_json, result)

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.create_reservation_provider')
    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get')
    @mock.patch('keystoneauth1.session.Session.request')
    def test_update_reservation_inventory_no_rp(
            self, kss_req, client_get, create_rp, get_rp):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"

        # Build the mock that there is no existing reservation provider
        get_rp.return_value = None

        # Build the mock of created resource provider
        mock_post_rp_json = {'uuid': rp_uuid,
                             'name': rp_name,
                             'generation': 0,
                             'parent_provider_uuid': host_uuid}
        create_rp.return_value = mock_post_rp_json

        # Build the mock of "current" inventory for get_inventory()
        curr_gen = 0
        mock_get_inv_json = {
            'inventories': {},
            "resource_provider_generation": curr_gen
        }
        client_get.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_get_inv_json))

        # Build the mock of "updated" inventory for update_inventory()
        update_gen = 1
        mock_put_json = {
            'inventories': {
                'CUSTOM_RESERVATION_ADD': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                },
            },
            "resource_provider_generation": update_gen
        }
        kss_req.return_value = fake_requests.FakeResponse(
            200, content=jsonutils.dump_as_bytes(mock_put_json))

        result = self.client.update_reservation_inventory(host_name, 'add', 3)

        # Ensure that the create_reservation_provider() is called.
        create_rp.assert_called_once_with(host_name)

        expected_data = {
            'inventories': {
                'CUSTOM_RESERVATION_ADD': {
                    "allocation_ratio": 1.0,
                    "max_unit": 1,
                    "min_unit": 1,
                    "reserved": 0,
                    "step_size": 1,
                    "total": 3
                },
            },
            "resource_provider_generation": curr_gen
        }
        expected_url = '/resource_providers/%s/inventories' % rp_uuid
        self._assert_keystone_called_once(kss_req, expected_url, 'PUT',
                                          json=expected_data)
        self.assertEqual(mock_put_json, result)

        kss_req.reset_mock()

        # Test retrying on 409 conflict
        mock_json_data = {
            "errors": [
                {"status": 409,
                 "code": "placement.concurrent_update",
                 "title": "Conflict"}
            ]
        }

        kss_req.return_value = fake_requests.FakeResponse(
            409, content=jsonutils.dump_as_bytes(mock_json_data))
        self.assertRaises(
            exceptions.InventoryConflict,
            self.client.update_reservation_inventory, host_name, 'add', 3)
        self.assertEqual(5, kss_req.call_count)
        kss_req.reset_mock()

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_inventory(self, kss_req, get_rp):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"

        # Build the mock of current resource provider
        mock_get_rp_json = {'uuid': rp_uuid,
                            'name': rp_name,
                            'generation': 0,
                            'parent_provider_uuid': host_uuid}
        get_rp.return_value = mock_get_rp_json

        kss_req.return_value = fake_requests.FakeResponse(200)

        self.client.delete_reservation_inventory(host_name, "curr1")

        expected_url = ('/resource_providers/%s/inventories'
                        '/CUSTOM_RESERVATION_CURR1' % rp_uuid)

        self._assert_keystone_called_once(kss_req, expected_url, 'DELETE')

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    def test_delete_reservation_inventory_no_rp(self, get_rp):
        host_name = "compute-1"
        # Build the mock that there is no existing reservation provider
        get_rp.return_value = None

        self.assertRaises(
            exceptions.ResourceProviderNotFound,
            self.client.delete_reservation_inventory, host_name, "curr1")

    @mock.patch('blazar.utils.openstack.placement.'
                'BlazarPlacementClient.get_resource_provider')
    @mock.patch('keystoneauth1.session.Session.request')
    def test_delete_reservation_inventory_no_rc(self, kss_req, get_rp):
        host_uuid = uuidutils.generate_uuid()
        host_name = "compute-1"
        rp_uuid = uuidutils.generate_uuid()
        rp_name = "blazar_compute-1"

        # Build the mock of current resource provider
        mock_get_rp_json = {'uuid': rp_uuid,
                            'name': rp_name,
                            'generation': 0,
                            'parent_provider_uuid': host_uuid}
        get_rp.return_value = mock_get_rp_json

        # If no reservation class found or if no inventory found,
        # then the placement API returns 404 error.
        kss_req.return_value = fake_requests.FakeResponse(404)
        # Ensure that no error is raised
        self.client.delete_reservation_inventory(host_name, "curr1")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import oslo_messaging

from blazar.monitor import notification_monitor
from blazar.plugins import base
from blazar import tests


class DummyMonitorPlugin(base.BaseMonitorPlugin):
    def is_notification_enabled(self):
        return True

    def get_notification_event_types(self):
        return []

    def get_notification_topics(self):
        return []

    def notification_callback(self, event_type, message):
        return {}

    def is_polling_enabled(self):
        return False

    def get_polling_interval(self):
        return 0

    def poll(self):
        return {}

    def get_healing_interval(self):
        return 0

    def heal(self):
        return {}


class NotificationMonitorTestCase(tests.TestCase):
    def setUp(self):
        super(NotificationMonitorTestCase, self).setUp()
        listener = self.patch(oslo_messaging, 'get_notification_listener')
        listener.return_value = None
        self.plugins = [DummyMonitorPlugin()]
        self.monitor = notification_monitor.NotificationMonitor(self.plugins)

    def test_get_targets(self):
        get_topics = self.patch(self.plugins[0], 'get_notification_topics')
        get_topics.return_value = ['topic1', 'topic2']

        targets = self.monitor._get_targets(self.plugins)
        self.assertEqual(2, len(targets))

    def test_get_endpoints(self):
        get_event_types = self.patch(self.plugins[0],
                                     'get_notification_event_types')
        get_event_types.return_value = ['event_type1', 'event_type2']

        endpoint = self.patch(notification_monitor.NotificationEndpoint,
                              '__init__')
        endpoint.return_value = None

        self.monitor._get_endpoints(self.plugins)
        endpoint.assert_called_once()


class NotificationEndpointTestCase(tests.TestCase):
    def setUp(self):
        super(NotificationEndpointTestCase, self).setUp()
        self.handler = self.patch(
            DummyMonitorPlugin, 'notification_callback'
        )
        plugin = DummyMonitorPlugin()
        monitor = notification_monitor.NotificationMonitor([plugin])
        monitor.handlers['event_type1'].append(plugin.notification_callback)
        self.endpoint = notification_monitor.NotificationEndpoint(monitor)

    def test_info(self):
        self.endpoint.info('dummy_ctxt', 'dummy_id', 'event_type1',
                           'hello', 'dummy_metadata')
        self.handler.assert_called_once_with('event_type1', 'hello')

    def test_warn(self):
        self.endpoint.warn('dummy_ctxt', 'dummy_id', 'event_type1',
                           'hello', 'dummy_metadata')
        self.handler.assert_called_once_with('event_type1', 'hello')

    def test_error(self):
        self.endpoint.error('dummy_ctxt', 'dummy_id', 'event_type1',
                            'hello', 'dummy_metadata')
        self.handler.assert_called_once_with('event_type1', 'hello')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging

from blazar.monitor import base

LOG = logging.getLogger(__name__)


class NotificationMonitor(base.BaseMonitor):
    """A notification based monitor."""

    def __init__(self, monitor_plugins):
        """Initialize a notification monitor."""
        LOG.debug('Initializing a notification monitor...')
        super(NotificationMonitor, self).__init__(monitor_plugins)
        try:
            self.handlers = defaultdict(list)
            self.listener = oslo_messaging.get_notification_listener(
                oslo_messaging.get_notification_transport(cfg.CONF),
                self._get_targets(monitor_plugins),
                self._get_endpoints(monitor_plugins),
                executor='eventlet'
            )
            LOG.debug('Notification listener is successfully created.')
        except Exception as e:
            LOG.exception('Failed to create a notification listener. (%s)',
                          str(e))

    def start_monitoring(self):
        """Start subscribing notifications."""
        LOG.debug('Starting a notification monitor...')
        try:
            self.listener.start()
            super(NotificationMonitor, self).start_monitoring()
        except Exception as e:
            LOG.exception('Failed to start a notification monitor. (%s)',
                          str(e))

    def stop_monitoring(self):
        """Stop subscribing notifications."""
        LOG.debug('Stopping a notification monitor...')
        try:
            self.listener.stop()
            super(NotificationMonitor, self).stop_monitoring()
        except Exception as e:
            LOG.exception('Failed to stop a notification monitor. (%s)',
                          str(e))

    def _get_targets(self, monitor_plugins):
        """Get a list of targets to subscribe.

        :param monitor_plugins: a list of resource monitor plugins. These
                                plugins provide notification topics.
        :return: a list of targets to subscribe.
        """
        topics = []
        for plugin in monitor_plugins:
            topics += plugin.get_notification_topics()

        return [oslo_messaging.Target(topic=topic) for topic in set(topics)]

    def _get_endpoints(self, monitor_plugins):
        """Get a list of endpoints which handle notification messages.

        :param monitor_plugins: a list of resource monitor plugins. These
                                plugins provide event types to handle and a
                                handler which is called when notification
                                messages of those event types
                                are received.
        :return: a list of endpoints.
        """
        for plugin in monitor_plugins:
            for event_type in plugin.get_notification_event_types():
                self.handlers[event_type].append(
                    # Wrap the notification callback with the
                    # call_monitor_plugin() to manage lease/reservation flags.
                    lambda e_type, payload: self.call_monitor_plugin(
                        plugin.notification_callback, e_type, payload))

        return [NotificationEndpoint(self)]


class NotificationEndpoint(object):
    """End point of notifications.

    It handles the notification messages. If it receives a notification
    message, it calls notification handlers provided as a parameter for the
    __init__().
    """

    def __init__(self, monitor, filter_rules={}):
        self.monitor = monitor
        self.filter_rule = oslo_messaging.NotificationFilter(**filter_rules)

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        self.handle_notification('INFO', ctxt, publisher_id, event_type,
                                 payload, metadata)

    def warn(self, ctxt, publisher_id, event_type, payload, metadata):
        self.handle_notification('WARN', ctxt, publisher_id, event_type,
                                 payload, metadata)

    def error(self, ctxt, publisher_id, event_type, payload, metadata):
        self.handle_notification('ERROR', ctxt, publisher_id, event_type,
                                 payload, metadata)

    def handle_notification(self, priority, ctxt, publisher_id, event_type,
                            payload, metadata):
        LOG.debug('Received a notification: priority: %s, publisher id: %s, '
                  'event type: %s, payload: %s', priority, publisher_id,
                  event_type, payload)
        for handler in self.monitor.handlers[str(event_type)]:
            handler(event_type, payload)



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add status to leases

Revision ID: 23d6240b51b2
Revises: 2bcfe76b0474
Create Date: 2014-04-25 10:41:09.183430

"""

# revision identifiers, used by Alembic.
revision = '23d6240b51b2'
down_revision = '2bcfe76b0474'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('leases', sa.Column(
        'action', sa.String(length=255), nullable=True))
    op.add_column('leases', sa.Column(
        'status', sa.String(length=255), nullable=True))
    op.add_column('leases', sa.Column(
        'status_reason', sa.String(length=255), nullable=True))


def downgrade():
    engine = op.get_bind().engine
    if engine.name == 'sqlite':
        # Only for testing purposes with sqlite
        op.execute('CREATE TABLE tmp_leases as SELECT created_at, updated_at, '
                   'id, name, user_id, project_id, start_date, '
                   'end_date, trust_id FROM leases')
        op.execute('DROP TABLE leases')
        op.execute('ALTER TABLE tmp_leases RENAME TO leases')
        return

    op.drop_column('leases', 'action')
    op.drop_column('leases', 'status')
    op.drop_column('leases', 'status_reason')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import opts


cli_opts = [
    cfg.HostAddressOpt('host', default='0.0.0.0',
                       help='Name of this node. This can be an opaque '
                            'identifier. It is not necessarily a hostname, '
                            'FQDN, or IP address. However, the node name must '
                            'be valid within an AMQP key, and if using '
                            'ZeroMQ (will be removed in the Stein release), a '
                            'valid hostname, FQDN, or IP address'),
    cfg.BoolOpt('log_exchange', default=False,
                help='Log request/response exchange details: environ, '
                     'headers and bodies'),
]

os_opts = [
    cfg.StrOpt('os_auth_protocol',
               default='http',
               help='Protocol used to access OpenStack Identity service'),
    cfg.HostAddressOpt('os_auth_host',
                       default='127.0.0.1',
                       help='IP or hostname of machine on which OpenStack '
                            'Identity service is located'),
    cfg.StrOpt('os_auth_port',
               default='5000',
               help='Port of OpenStack Identity service.'),
    cfg.StrOpt('os_auth_prefix',
               default='',
               help='Prefix of URL to access OpenStack Identity service.'),
    cfg.StrOpt('os_admin_username',
               default='admin',
               help='This OpenStack user is used to treat trusts. '
                    'The user must have admin role in <os_admin_project_name> '
                    'project.'),
    cfg.StrOpt('os_admin_password',
               default='blazar',
               help='Password of the admin user to treat trusts.'),
    cfg.StrOpt('os_admin_project_name',
               default='admin',
               help='Name of project where the user is admin.'),
    cfg.StrOpt('os_auth_version',
               default='v3',
               help='Blazar uses API v3 to allow trusts using.'),
    cfg.StrOpt('os_admin_user_domain_name',
               default='Default',
               help='A domain name the os_admin_username belongs to.'),
    cfg.StrOpt('os_admin_project_domain_name',
               default='Default',
               help='A domain name the os_admin_project_name belongs to'),
    cfg.StrOpt('cafile',
               help='Path of the custom CA certificates bundle.'),
]

api_opts = [
    cfg.BoolOpt('enable_v1_api',
                default=True,
                help='Deploy the v1 API.'),
]

lease_opts = [
    cfg.IntOpt('cleaning_time',
               default=0,
               min=0,
               help='The minimum interval [minutes] between the end of a '
                    'lease and the start of the next lease for the same '
                    'resource. This interval is used for cleanup.')
]

CONF = cfg.CONF
CONF.register_cli_opts(cli_opts)
CONF.register_opts(os_opts)
CONF.register_opts(api_opts)
CONF.register_opts(lease_opts)
logging.register_options(cfg.CONF)


def set_lib_defaults():
    """Update default value for configuration options from other namespace.

    Example, oslo lib config options. This is needed for
    config generator tool to pick these default value changes.
    https://docs.openstack.org/oslo.config/latest/cli/
    generator.html#modifying-defaults-from-other-namespaces
    """

    # TODO(gmann): Remove setting the default value of config policy_file
    # once oslo_policy change the default value to 'policy.yaml'.
    # https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
    opts.set_defaults(CONF, 'policy.yaml')



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import operator

from oslo_utils import uuidutils

from blazar.db import exceptions as db_exceptions
from blazar.db.sqlalchemy import api as db_api
from blazar.plugins import oshosts as host_plugin
from blazar import tests


def _get_fake_random_uuid():
    return uuidutils.generate_uuid()


def _get_fake_lease_uuid():
    """Returns a fake uuid."""
    return 'aaaaaaaa-1111-bbbb-2222-cccccccccccc'


def _get_fake_phys_reservation_values(id=None,
                                      lease_id=_get_fake_lease_uuid(),
                                      resource_id=None):
    if id is None:
        id = _get_fake_random_uuid()
    return {'id': id,
            'lease_id': lease_id,
            'resource_id': '1234' if not resource_id else resource_id,
            'resource_type': host_plugin.RESOURCE_TYPE,
            'hypervisor_properties': '[\"=\", \"$hypervisor_type\", \"QEMU\"]',
            'resource_properties': '',
            'min': 1, 'max': 1,
            'trust_id': 'exxee111qwwwwe'}


def _get_fake_event_values(id=None,
                           lease_id=_get_fake_lease_uuid(),
                           event_type='fake_event_type',
                           time=None,
                           status='fake_event_status'):
    if id is None:
        id = _get_fake_random_uuid()
    return {'id': id,
            'lease_id': lease_id,
            'event_type': event_type,
            'time': _get_datetime('2030-03-01 00:00') if not time else time,
            'status': status}


def _get_datetime(value='2030-01-01 00:00'):
    return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M')


def _get_fake_phys_lease_values(id=None,
                                name='fake_phys_lease',
                                start_date=_get_datetime('2030-01-01 00:00'),
                                end_date=_get_datetime('2030-01-02 00:00'),
                                resource_id=None):
    if id is None:
        id = _get_fake_random_uuid()
    return {'id': id,
            'name': name,
            'user_id': 'fake',
            'project_id': 'fake',
            'start_date': start_date,
            'end_date': end_date,
            'trust': 'trust',
            'reservations': [_get_fake_phys_reservation_values(
                id=_get_fake_random_uuid(),
                lease_id=id,
                resource_id=resource_id)],
            'events': []
            }


def _get_fake_host_allocation_values(
        id=None,
        compute_host_id=_get_fake_random_uuid(),
        reservation_id=_get_fake_random_uuid()):
    values = {'compute_host_id': compute_host_id,
              'reservation_id': reservation_id}
    if id is not None:
        values.update({'id': id})

    return values


def _create_physical_lease(values=_get_fake_phys_lease_values(),
                           random=False):
    """Creating fake lease having a single physical resource."""
    if random is True:
        values = _get_fake_phys_lease_values(id=_get_fake_random_uuid(),
                                             name=_get_fake_random_uuid())
    lease = db_api.lease_create(values)
    phys_res = _get_fake_phys_reservation_values()
    for reservation in db_api.reservation_get_all_by_lease_id(lease['id']):
        allocation_values = {
            'id': _get_fake_random_uuid(),
            'compute_host_id': values['reservations'][0]['resource_id'],
            'reservation_id': reservation['id']
        }
        db_api.host_allocation_create(allocation_values)
        computehost_reservation = {
            'id': _get_fake_random_uuid(),
            'reservation_id': values['reservations'][0]['id'],
            'resource_properties': phys_res['resource_properties'],
            'hypervisor_properties': phys_res['hypervisor_properties'],
            'count_range': "{0} - {1}".format(phys_res['min'],
                                              phys_res['max'])
        }
        db_api.host_reservation_create(computehost_reservation)
    return lease


def _get_fake_host_reservation_values(id=None, reservation_id=None):
    if id is None:
        id = _get_fake_random_uuid()
    if reservation_id is None:
        reservation_id = _get_fake_random_uuid()
    return {'id': id,
            'reservation_id': reservation_id,
            'resource_properties': "fake",
            'hypervisor_properties': "fake",
            'min': 1, 'max': 1,
            'trust_id': 'exxee111qwwwwe'}


def _get_fake_instance_values(id=None, reservation_id=None):
    if id is None:
        id = _get_fake_random_uuid()
    if reservation_id is None:
        reservation_id = _get_fake_random_uuid()
    return {'id': id,
            'reservation_id': reservation_id,
            'vcpus': 1,
            'memory_mb': 2024,
            'disk_gb': 100,
            'amount': 2,
            'affinity': False,
            'flavor_id': 'fake_flavor_id',
            'aggregate_id': 29,
            'server_group_id': 'server_group_id'}


def _get_fake_cpu_info():
    return str({'vendor': 'Intel',
                'model': 'Westmere',
                'arch': 'x86_64',
                'features': ['rdtscp', 'pdpe1gb', 'hypervisor', 'vmx', 'ss',
                             'vme'],
                'topology': {'cores': 1, 'threads': 1, 'sockets': 2}})


def _get_fake_host_values(id=None, mem=8192, disk=10):
    if id is None:
        id = _get_fake_random_uuid()
    return {'id': id,
            'availability_zone': 'az1',
            'vcpus': 1,
            'cpu_info': _get_fake_cpu_info(),
            'hypervisor_type': 'QEMU',
            'hypervisor_version': 1000,
            'memory_mb': mem,
            'local_gb': disk,
            'status': 'free',
            'trust_id': 'exxee111qwwwwe',
            }


def _get_fake_host_extra_capabilities(id=None,
                                      computehost_id=None,
                                      name='vgpu',
                                      value='2'):
    if id is None:
        id = _get_fake_random_uuid()
    if computehost_id is None:
        computehost_id = _get_fake_random_uuid()
    return {'id': id,
            'computehost_id': computehost_id,
            'property_name': name,
            'capability_value': value}


def is_result_sorted_correctly(results, sort_key, sort_dir='asc'):
    sorted_list = sorted(results,
                         key=operator.itemgetter(sort_key),
                         reverse=False if sort_dir == 'asc' else True)
    return sorted_list == results


class SQLAlchemyDBApiTestCase(tests.DBTestCase):
    """Test case for SQLAlchemy DB API."""

    def setUp(self):
        super(SQLAlchemyDBApiTestCase, self).setUp()

    def test_create_phys_lease(self):
        """Check physical lease create

        Create a physical lease and verify that all tables have been
        populated.
        """

        result = db_api.lease_create(_get_fake_phys_lease_values())
        self.assertEqual(result['name'],
                         _get_fake_phys_lease_values()['name'])
        self.assertEqual(0, len(db_api.event_get_all()))
        self.assertEqual(1, len(db_api.reservation_get_all()))

    def test_create_duplicate_leases(self):
        """Create two leases with same ids, and checks it raises an error."""

        db_api.lease_create(_get_fake_phys_lease_values(id='42'))
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.lease_create,
                          _get_fake_phys_lease_values(id='42'))

    def test_create_leases_with_duplicated_reservation(self):
        """Check duplicated reservation create

        Create two leases with a duplicated reservation,
        and checks it raises an error.
        """
        lease_values = _get_fake_phys_lease_values()

        db_api.lease_create(lease_values)

        lease_values['id'] = _get_fake_random_uuid()
        lease_values['name'] = 'duplicated_reservation'

        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.lease_create,
                          lease_values)

    def test_create_leases_with_duplicated_event(self):
        """Check duplicated event create

        Create two leases with a duplicated event,
        and checks it raises an error.
        """
        lease_values = _get_fake_phys_lease_values()
        lease_values['events'] = [_get_fake_event_values()]

        db_api.lease_create(lease_values)

        lease_values['id'] = _get_fake_random_uuid()
        lease_values['name'] = 'duplicated_event'
        lease_values['reservations'][0]['id'] = _get_fake_random_uuid()

        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.lease_create,
                          lease_values)

    def test_create_lease_with_event(self):
        """Create a lease including a fake event and check all tables."""

        lease = _get_fake_phys_lease_values()
        lease['events'].append(_get_fake_event_values(lease_id=lease['id']))
        result = db_api.lease_create(lease)
        self.assertEqual(result['name'],
                         _get_fake_phys_lease_values()['name'])
        self.assertEqual(1, len(db_api.event_get_all()))

    def test_delete_wrong_lease(self):
        """Delete a lease that doesn't exist and check that raises an error."""
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.lease_destroy, 'fake_id')

    def test_get_physical_lease(self):
        """Test if physical host reservation contains data of reservation."""
        lease = _get_fake_phys_lease_values()
        lease['events'].append(_get_fake_event_values(lease_id=lease['id']))
        result = _create_physical_lease(values=lease)
        result = db_api.lease_get(result['id'])
        res = result.to_dict()
        self.assertEqual(res['reservations'][0]['hypervisor_properties'],
                         lease['reservations'][0]['hypervisor_properties'])
        self.assertEqual(res['reservations'][0]['resource_properties'],
                         lease['reservations'][0]['resource_properties'])
        self.assertEqual(res['reservations'][0]['min'],
                         lease['reservations'][0]['min'])
        self.assertEqual(res['reservations'][0]['max'],
                         lease['reservations'][0]['max'])

    def test_delete_correct_lease(self):
        """Delete a lease and check that deletion has been cascaded to FKs."""
        lease = _get_fake_phys_lease_values()
        lease['events'].append(_get_fake_event_values(lease_id=lease['id']))
        result = _create_physical_lease(values=lease)
        db_api.lease_destroy(result['id'])
        self.assertIsNone(db_api.lease_get(result['id']))
        self.assertEqual(0, len(db_api.reservation_get_all()))
        self.assertEqual(0, len(db_api.event_get_all()))

    def test_lease_get_all(self):
        """Check the number of leases we get."""
        _create_physical_lease(random=True)
        self.assertEqual(1, len(db_api.lease_get_all()))
        _create_physical_lease(random=True)
        self.assertEqual(2, len(db_api.lease_get_all()))

    def test_lease_list(self):
        """Not implemented yet until lease_list returns list of IDs."""
        # TODO(sbauza): Enable this test when lease_list will return only IDs
        self.assertTrue(True)
        return
        _create_physical_lease(
            values=_get_fake_phys_lease_values(id='1', name='fake1'))
        _create_physical_lease(
            values=_get_fake_phys_lease_values(id='2', name='fake2'))
        self.assertEqual(['1', '2'], db_api.lease_list())

    def test_lease_update(self):
        """Update both start_date and name and check lease has been updated."""
        result = _create_physical_lease()
        result = db_api.lease_update(result['id'],
                                     values={'name': 'lease_renamed'})
        self.assertEqual('lease_renamed', result['name'])
        result = db_api.lease_update(
            result['id'],
            values={'start_date': _get_datetime('2014-02-01 00:00')})
        self.assertEqual(_get_datetime('2014-02-01 00:00'),
                         result['start_date'])

    # Reservations

    def test_create_reservation(self):
        """Create and verify reservation

        Create a reservation and verify that all tables have been
        populated.
        """

        result = db_api.reservation_create(_get_fake_phys_reservation_values())
        self.assertEqual(result['lease_id'],
                         _get_fake_phys_reservation_values()
                         ['lease_id'])

    def test_reservation_get_all_by_values(self):
        """Create 2 reservations and check find abilities

        Create two reservations and verify that we can find reservation per
        resource_id or resource_type.
        """
        db_api.reservation_create(
            _get_fake_phys_reservation_values(id='1', resource_id='1234'))
        db_api.reservation_create(
            _get_fake_phys_reservation_values(id='2', resource_id='5678'))
        self.assertEqual(2, len(db_api.reservation_get_all_by_values()))
        self.assertEqual(1, len(db_api.reservation_get_all_by_values(
            resource_id='5678')))
        self.assertEqual(2, len(db_api.reservation_get_all_by_values(
            resource_type=host_plugin.RESOURCE_TYPE)))

    def test_reservation_update(self):
        result = db_api.reservation_create(_get_fake_phys_reservation_values())
        self.assertNotEqual('fake', result.resource_type)

        result = db_api.reservation_update(result.id,
                                           {"resource_type": 'fake'})
        self.assertEqual('fake', result.resource_type)

    def test_reservation_destroy_for_reservation_not_found(self):
        self.assertFalse(db_api.reservation_get('1'))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.reservation_destroy, '1')

    def test_create_duplicate_reservation(self):
        """Create duplicated reservation

        Create a reservation and verify that an exception is raised if a
        duplicated reservation is created.
        """
        uuid = _get_fake_random_uuid()
        db_api.reservation_create(_get_fake_phys_reservation_values(id=uuid))
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.reservation_create,
                          _get_fake_phys_reservation_values(id=uuid))

    # Host reservations

    def test_create_host_reservation(self):
        """Create host reservation

        Create a host reservation and verify that all tables
        have been populated.
        """

        result = db_api.host_reservation_create(
            _get_fake_host_reservation_values(id='1'))
        self.assertEqual(result['id'],
                         _get_fake_host_reservation_values(id='1')
                         ['id'])

    def test_create_duplicate_host_reservation(self):
        """Create duplicated host reservation

        Create a duplicated host reservation and verify that an exception is
        raised.
        """

        db_api.host_reservation_create(
            _get_fake_host_reservation_values(id='1'))
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.host_reservation_create,
                          _get_fake_host_reservation_values(id='1'))

    def test_delete_host_reservation(self):
        """Check deletion for host reservation

        Check all deletion cases for host reservation,
        including cascade deletion from reservations table.
        """

        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_reservation_destroy, 'fake_id')

        result = db_api.host_reservation_create(
            _get_fake_host_reservation_values())
        db_api.host_reservation_destroy(result['id'])
        self.assertIsNone(db_api.host_reservation_get(result['id']))
        reserv = db_api.reservation_create(_get_fake_phys_reservation_values())
        result = db_api.host_reservation_create(
            _get_fake_host_reservation_values(reservation_id=reserv['id']))
        db_api.reservation_destroy(reserv['id'])
        self.assertIsNone(db_api.host_reservation_get(result['id']))

    def test_host_reservation_get_all(self):
        """Check that we return 2 hosts."""

        db_api.host_reservation_create(_get_fake_host_reservation_values(id=1))
        db_api.host_reservation_create(_get_fake_host_reservation_values(id=2))
        hosts_reservations = db_api.host_reservation_get_all()
        self.assertEqual(['1', '2'], [x['id'] for x in hosts_reservations])

    def test_host_reservation_get_by_reservation_id(self):
        """Check that we return 2 hosts."""

        db_api.host_reservation_create(
            _get_fake_host_reservation_values(id=1, reservation_id=1))
        db_api.host_reservation_create(
            _get_fake_host_reservation_values(id=2, reservation_id=2))
        res = db_api.host_reservation_get_by_reservation_id(2)
        self.assertEqual('2', res['id'])

    def test_update_host_reservation(self):
        db_api.host_reservation_create(_get_fake_host_reservation_values(id=1))
        db_api.host_reservation_update(1, {'resource_properties': 'updated'})
        res = db_api.host_reservation_get(1)
        self.assertEqual('updated', res['resource_properties'])

    def test_create_host(self):
        """Create a host and verify that all tables have been populated."""
        result = db_api.host_create(_get_fake_host_values(id='1'))
        self.assertEqual(result['id'], _get_fake_host_values(id='1')['id'])

    def test_create_duplicated_host(self):
        """Create a duplicated host and verify that an exception is raised."""
        db_api.host_create(_get_fake_host_values(id='1'))
        # Making sure we still raise a DuplicateDBEntry
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.host_create,
                          _get_fake_host_values(id='1'))

    def test_search_for_hosts_by_ram(self):
        """Check RAM info search

        Create two hosts and check that we can find a host per its RAM
        info.
        """
        db_api.host_create(_get_fake_host_values(id=1, mem=2048))
        db_api.host_create(_get_fake_host_values(id=2, mem=4096))
        self.assertEqual(2, len(
            db_api.host_get_all_by_queries(['memory_mb >= 2048'])))
        self.assertEqual(0, len(
            db_api.host_get_all_by_queries(['memory_mb lt 2048'])))

    def test_search_for_hosts_by_cpu_info(self):
        """Create one host and search within cpu_info."""

        db_api.host_create(_get_fake_host_values())
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['cpu_info like %Westmere%'])))

    def test_search_for_hosts_by_extra_capability(self):
        """Create one host and test extra capability queries."""
        # We create a first host, with extra capabilities
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.resource_property_create(dict(
            id='a', resource_type='physical:host', private=False,
            property_name='vgpu'))
        db_api.resource_property_create(dict(
            id='b', resource_type='physical:host', private=False,
            property_name='nic_model'))
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(computehost_id=1))
        db_api.host_extra_capability_create(_get_fake_host_extra_capabilities(
            computehost_id=1,
            name='nic_model',
            value='ACME Model A',
        ))
        # We create a second host, without any extra capabilities
        db_api.host_create(_get_fake_host_values(id=2))

        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['vgpu == 2'])))
        self.assertEqual(0, len(
            db_api.host_get_all_by_queries(['vgpu != 2'])))
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['cpu_info like %Westmere%',
                                            'vgpu == 2'])))
        self.assertEqual(0, len(
            db_api.host_get_all_by_queries(['cpu_info like %wrongcpu%',
                                            'vgpu == 2'])))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_get_all_by_queries, ['apples < 2048'])
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['nic_model == ACME Model A'])
        ))

    def test_resource_properties_list(self):
        """Create one host and test extra capability queries."""
        # We create a first host, with extra capabilities
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.resource_property_create(dict(
            id='a', resource_type='physical:host', private=False,
            property_name='vgpu'))
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(computehost_id=1))

        result = db_api.resource_properties_list('physical:host')

        self.assertListEqual(result, [('vgpu', False, '2')])

    def test_search_for_hosts_by_composed_queries(self):
        """Create one host and test composed queries."""

        db_api.host_create(_get_fake_host_values(mem=8192))
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['memory_mb > 2048',
                                            'cpu_info like %Westmere%'])))
        self.assertEqual(0, len(
            db_api.host_get_all_by_queries(['memory_mb < 2048',
                                            'cpu_info like %Westmere%'])))
        self.assertRaises(db_exceptions.BlazarDBInvalidFilter,
                          db_api.host_get_all_by_queries, ['memory_mb <'])
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_get_all_by_queries, ['apples < 2048'])
        self.assertRaises(db_exceptions.BlazarDBInvalidFilterOperator,
                          db_api.host_get_all_by_queries,
                          ['memory_mb wrongop 2048'])
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['memory_mb in 4096,8192'])))
        self.assertEqual(1, len(
            db_api.host_get_all_by_queries(['memory_mb != null'])))

    def test_list_hosts(self):
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.host_create(_get_fake_host_values(id=2))
        self.assertEqual(2, len(db_api.host_list()))

    def test_get_hosts_per_filter(self):
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.host_create(_get_fake_host_values(id=2))
        filters = {'status': 'free'}
        self.assertEqual(2, len(
            db_api.host_get_all_by_filters(filters)))

    def test_update_host(self):
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.host_update(1, {'status': 'updated'})
        self.assertEqual('updated', db_api.host_get(1)['status'])

    def test_delete_host(self):
        db_api.host_create(_get_fake_host_values(id=1))
        db_api.host_destroy(1)
        self.assertIsNone(db_api.host_get(1))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_destroy, 2)

    def test_create_host_extra_capability(self):
        db_api.resource_property_create(dict(
            id='id', resource_type='physical:host', private=False,
            property_name='vgpu'))
        result, _ = db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id=1, name='vgpu'))

        self.assertEqual(result.id, _get_fake_host_values(id='1')['id'])

    def test_create_duplicated_host_extra_capability(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id=1))
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.host_extra_capability_create,
                          _get_fake_host_extra_capabilities(id='1'))

    def test_get_host_extra_capability_per_id(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='1'))
        result, _ = db_api.host_extra_capability_get('1')
        self.assertEqual('1', result.id)

    def test_host_extra_capability_get_all_per_host(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='1', computehost_id='1'))
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='2', computehost_id='1'))
        res = db_api.host_extra_capability_get_all_per_host('1')
        self.assertEqual(2, len(res))

    def test_update_host_extra_capability(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='1'))
        db_api.host_extra_capability_update('1', {'capability_value': '2'})
        res, _ = db_api.host_extra_capability_get('1')
        self.assertEqual('2', res.capability_value)

    def test_delete_host_extra_capability(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='1'))
        db_api.host_extra_capability_destroy('1')
        self.assertIsNone(db_api.host_extra_capability_get('1'))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_extra_capability_destroy, '1')

    def test_host_extra_capability_get_all_per_name(self):
        db_api.host_extra_capability_create(
            _get_fake_host_extra_capabilities(id='1', computehost_id='1'))
        res = db_api.host_extra_capability_get_all_per_name('1', 'vgpu')
        self.assertEqual(1, len(res))
        self.assertEqual([],
                         db_api.host_extra_capability_get_all_per_name('1',
                                                                       'bad'))

    # Resource Inventory

    def test_host_resource_inventory_create(self):
        db_api.host_create(_get_fake_host_values(id=1))
        fake_inventory_values = {
            'id': _get_fake_random_uuid(),
            'computehost_id': '1',
            'resource_class': 'VCPU',
            'total': 10,
            'reserved': 2,
            'min_unit': 1,
            'max_unit': 1,
            'step_size': 1,
            'allocation_ratio': 1.0
        }

        db_api.host_resource_inventory_create(fake_inventory_values)

        host = db_api.host_get(1)
        actual = host.computehost_resource_inventories[0].to_dict()
        self.assertIsNotNone(actual["created_at"])
        del actual["created_at"]
        expected = fake_inventory_values.copy()
        expected["updated_at"] = None
        self.assertDictEqual(expected, actual)

    def test_host_resource_inventory_get_all_per_host(self):
        db_api.host_create(_get_fake_host_values(id=1))
        fake_inventory_values = {
            'id': _get_fake_random_uuid(),
            'computehost_id': '1',
            'resource_class': 'VCPU',
            'total': 10,
            'reserved': 2,
            'min_unit': 1,
            'max_unit': 1,
            'step_size': 1,
            'allocation_ratio': 1.0
        }
        db_api.host_resource_inventory_create(fake_inventory_values)

        result = db_api.host_resource_inventory_get_all_per_host("1")

        result = list(result)
        self.assertEqual(1, len(result))
        first_result = result[0].to_dict()
        expected = fake_inventory_values.copy()
        expected["updated_at"] = None
        expected["created_at"] = first_result["created_at"]
        self.assertDictEqual(expected, first_result)

    # Trait

    def test_host_traits_create(self):
        db_api.host_create(_get_fake_host_values(id=1))
        fake_trait_values = {
            'id': _get_fake_random_uuid(),
            'computehost_id': '1',
            'trait': 'HW_CPU_X86_AVX'
        }

        db_api.host_trait_create(fake_trait_values)

        host = db_api.host_get(1)
        actual = host.computehost_traits[0].to_dict()
        self.assertIsNotNone(actual["created_at"])
        del actual["created_at"]
        expected = fake_trait_values.copy()
        expected["updated_at"] = None
        self.assertDictEqual(expected, actual)

    # Instance reservation

    def check_instance_reservation_values(self, expected, reservation_id):
        inst_reservation = db_api.instance_reservation_get(reservation_id)
        for k, v in expected.items():
            self.assertEqual(v, inst_reservation[k])

    def test_instance_reservation_create(self):
        reservation_values = _get_fake_instance_values(id='1')
        ret = db_api.instance_reservation_create(reservation_values)

        self.assertEqual('1', ret['id'])
        self.check_instance_reservation_values(reservation_values, '1')

    def test_create_duplicated_instance_reservation(self):
        reservation_values = _get_fake_instance_values(id='1')
        db_api.instance_reservation_create(reservation_values)
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.instance_reservation_create,
                          reservation_values)

    def test_instance_reservation_get(self):
        reservation1_values = _get_fake_instance_values(id='1')
        db_api.instance_reservation_create(reservation1_values)
        reservation2_values = _get_fake_instance_values(id='2')
        db_api.instance_reservation_create(reservation2_values)

        self.check_instance_reservation_values(reservation1_values, '1')
        self.check_instance_reservation_values(reservation2_values, '2')

    def test_instance_reservation_update(self):
        reservation_values = _get_fake_instance_values(id='1')
        db_api.instance_reservation_create(reservation_values)

        self.check_instance_reservation_values(reservation_values, '1')

        updated_values = {
            'flavor_id': 'updated-flavor-id',
            'aggregate_id': 30,
            'server_group_id': 'updated-server-group-id'
            }
        db_api.instance_reservation_update('1', updated_values)
        reservation_values.update(updated_values)
        self.check_instance_reservation_values(reservation_values, '1')

    def test_update_non_existing_instance_reservation(self):
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.instance_reservation_destroy, 'non-exists')

    def test_instance_reservation_destroy(self):
        reservation_values = _get_fake_instance_values(id='1')
        db_api.instance_reservation_create(reservation_values)

        self.check_instance_reservation_values(reservation_values, '1')

        db_api.instance_reservation_destroy('1')
        self.assertIsNone(db_api.instance_reservation_get('1'))

    def test_destroy_non_existing_instance_reservation(self):
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.instance_reservation_destroy, 'non-exists')

    # Host allocations

    def test_host_allocation_get_all(self):
        self.assertFalse(db_api.host_allocation_get_all())

        db_api.host_allocation_create(_get_fake_host_allocation_values(id='1'))
        db_api.host_allocation_create(_get_fake_host_allocation_values(id='2'))

        self.assertEqual(2, len(db_api.host_allocation_get_all()))

    def test_host_allocation_create_for_duplicated_hosts(self):
        db_api.host_allocation_create(
            _get_fake_host_allocation_values(id='1')
        )

        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.host_allocation_create,
                          _get_fake_host_allocation_values(id='1'))

    def test_host_allocation_update_for_host(self):
        host_allocation = db_api.host_allocation_create(
            _get_fake_host_allocation_values(
                compute_host_id="1",
                reservation_id="1"
            ))

        new_host_allocation = db_api.host_allocation_update(
            host_allocation.id,
            _get_fake_host_allocation_values(
                compute_host_id="2",
                reservation_id="2"
            ))

        self.assertEqual('2', new_host_allocation.compute_host_id)
        self.assertEqual('2', new_host_allocation.reservation_id)
        self.assertNotEqual(host_allocation.compute_host_id,
                            new_host_allocation.compute_host_id)

    def test_host_allocation_destroy_for_host(self):
        host_allocation = db_api.host_allocation_create(
            _get_fake_host_allocation_values()
        )
        db_api.host_allocation_destroy(host_allocation.id)

        self.assertIsNone(db_api.host_allocation_get(host_allocation.id))

    def test_host_allocation_destroy_for_host_not_found(self):
        host_allocation_id = _get_fake_random_uuid()

        self.assertIsNone(db_api.host_allocation_get(host_allocation_id))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.host_allocation_destroy,
                          host_allocation_id)

    def test_host_allocation_get_all_by_values(self):
        db_api.host_allocation_create(_get_fake_host_allocation_values(
            compute_host_id="1", reservation_id="1"))
        db_api.host_allocation_create(_get_fake_host_allocation_values(
            compute_host_id="1", reservation_id="1234"))

        self.assertEqual(2, len(db_api.host_allocation_get_all_by_values()))
        self.assertEqual(1, len(db_api.host_allocation_get_all_by_values(
            reservation_id='1234')))

    # Event

    def test_event_create(self):
        fake_event_type = 'test_event'

        test_event = db_api.event_create(_get_fake_event_values(
            event_type=fake_event_type))
        self.assertTrue(test_event)
        self.assertEqual(fake_event_type, test_event.event_type)

    def test_create_duplicated_event(self):
        self.assertFalse(db_api.event_get('1'))

        fake_values = _get_fake_event_values(id='1')
        test_event = db_api.event_create(fake_values)

        self.assertTrue(test_event)
        self.assertRaises(db_exceptions.BlazarDBDuplicateEntry,
                          db_api.event_create, fake_values)

    def test_event_update(self):
        self.assertFalse(db_api.event_get('1'))

        test_event = db_api.event_create(_get_fake_event_values(id='1'))

        self.assertTrue(test_event)

        test_event = db_api.event_update(test_event.id, {'status': 'changed'})

        self.assertEqual('changed', test_event.status)

    def test_event_destroy(self):
        self.assertFalse(db_api.event_get('1'))

        db_api.event_create(_get_fake_event_values(
            id='1'))
        self.assertTrue(db_api.event_get('1'))
        db_api.event_destroy('1')
        self.assertFalse(db_api.event_get('1'))

    def test_destroy_for_event_not_found(self):
        self.assertFalse(db_api.event_get('1'))
        self.assertRaises(db_exceptions.BlazarDBNotFound,
                          db_api.event_destroy, '1')

    def test_event_get_first_sorted_by_event_type_filter(self):
        fake_event_type = 'test_event'

        db_api.event_create(_get_fake_event_values(
            id='1'
        ))
        db_api.event_create(_get_fake_event_values(
            id='2',
            event_type=fake_event_type
        ))
        db_api.event_create(_get_fake_event_values(
            id='3',
            event_type=fake_event_type
        ))

        filtered_events = db_api.event_get_first_sorted_by_filters(
            sort_key='time',
            sort_dir='asc',
            filters={'event_type': fake_event_type}
        )
        self.assertEqual(fake_event_type, filtered_events.event_type)
        self.assertEqual('2', filtered_events.id)

    def test_event_get_first_sorted_by_status_filter(self):
        fake_status = 'test_status'
        db_api.event_create(_get_fake_event_values(
            id='1'
        ))
        db_api.event_create(_get_fake_event_values(
            id='2',
            status=fake_status
        ))
        db_api.event_create(_get_fake_event_values(
            id='3',
            status=fake_status
        ))

        filtered_events = db_api.event_get_first_sorted_by_filters(
            sort_key='time',
            sort_dir='asc',
            filters={'status': fake_status}
        )
        self.assertEqual(fake_status, filtered_events.status)
        self.assertEqual('2', filtered_events.id)

    def test_event_get_first_sorted_by_lease_id_filter(self):
        fake_lease_id = '1234'
        db_api.event_create(_get_fake_event_values(
            id='1'
        ))
        db_api.event_create(_get_fake_event_values(
            id='2',
            lease_id=fake_lease_id
        ))
        db_api.event_create(_get_fake_event_values(
            id='3',
            lease_id=fake_lease_id
        ))

        filtered_events = db_api.event_get_first_sorted_by_filters(
            sort_key='time',
            sort_dir='asc',
            filters={'lease_id': fake_lease_id}
        )
        self.assertEqual(fake_lease_id, filtered_events.lease_id)
        self.assertEqual('2', filtered_events.id)

    def test_event_get_sorted_asc_by_event_type_filter(self):
        fake_event_type = 'test_event'
        sort_dir = 'asc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            event_type=fake_event_type,
            time=datetime.datetime.utcnow()
        ))
        db_api.event_create(_get_fake_event_values(
            id='2',
            event_type=fake_event_type,
            time=datetime.datetime.utcnow()
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'event_type': fake_event_type}
        )
        self.assertEqual(2, len(filtered_events))
        self.assertEqual(fake_event_type, filtered_events[0].event_type)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_asc_by_status_filter(self):
        fake_status = 'test_status'
        sort_dir = 'asc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            status=fake_status
        ))
        db_api.event_create(_get_fake_event_values(
            id='2'
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'status': fake_status}
        )
        self.assertEqual(1, len(filtered_events))
        self.assertEqual(fake_status, filtered_events[0].status)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_asc_by_lease_id_filter(self):
        fake_lease_id = '1234'
        sort_dir = 'asc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            lease_id=fake_lease_id

        ))
        db_api.event_create(_get_fake_event_values(
            id='2'
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'lease_id': fake_lease_id}
        )
        self.assertEqual(1, len(filtered_events))
        self.assertEqual(fake_lease_id, filtered_events[0].lease_id)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_asc_by_time_filter(self):
        def check_query(border, op, expected_ids):
            filtered_events = db_api.event_get_all_sorted_by_filters(
                sort_key=sort_key,
                sort_dir=sort_dir,
                filters={'time': {'border': _get_datetime(border),
                                  'op': op}})
            filtered_event_ids = [e.id for e in filtered_events]
            self.assertListEqual(expected_ids, filtered_event_ids)

        time1 = _get_datetime('2030-01-01 01:00')
        time2 = _get_datetime('2030-01-01 02:00')
        time3 = _get_datetime('2030-01-01 03:00')
        sort_key = 'time'
        sort_dir = 'asc'

        db_api.event_create(_get_fake_event_values(id='1', time=time1))
        db_api.event_create(_get_fake_event_values(id='2', time=time2))
        db_api.event_create(_get_fake_event_values(id='3', time=time3))

        check_query('2030-01-01 02:00', 'lt', ['1'])
        check_query('2030-01-01 02:00', 'le', ['1', '2'])
        check_query('2030-01-01 02:00', 'gt', ['3'])
        check_query('2030-01-01 02:00', 'ge', ['2', '3'])
        check_query('2030-01-01 02:00', 'eq', ['2'])

    def test_event_get_sorted_desc_by_event_type_filter(self):
        fake_event_type = 'test_event'
        sort_dir = 'desc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            event_type=fake_event_type
        ))
        db_api.event_create(_get_fake_event_values(
            id='2',
            event_type=fake_event_type
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'event_type': fake_event_type}
        )
        self.assertEqual(2, len(filtered_events))
        self.assertEqual(fake_event_type, filtered_events[0].event_type)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_desc_by_status_filter(self):
        fake_status = 'test_status'
        sort_dir = 'desc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            status=fake_status
        ))
        db_api.event_create(_get_fake_event_values(
            id='2'
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'status': fake_status}
        )
        self.assertEqual(1, len(filtered_events))
        self.assertEqual(fake_status, filtered_events[0].status)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_desc_by_lease_id_filter(self):
        fake_lease_id = '1234'
        sort_dir = 'desc'
        sort_key = 'time'

        db_api.event_create(_get_fake_event_values(
            id='1',
            lease_id=fake_lease_id

        ))
        db_api.event_create(_get_fake_event_values(
            id='2'
        ))

        filtered_events = db_api.event_get_all_sorted_by_filters(
            sort_key=sort_key,
            sort_dir=sort_dir,
            filters={'lease_id': fake_lease_id}
        )
        self.assertEqual(1, len(filtered_events))
        self.assertEqual(fake_lease_id, filtered_events[0].lease_id)

        # testing sort
        self.assertTrue(is_result_sorted_correctly(filtered_events,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir))

    def test_event_get_sorted_desc_by_time_filter(self):
        def check_query(border, op, expected_ids):
            filtered_events = db_api.event_get_all_sorted_by_filters(
                sort_key=sort_key,
                sort_dir=sort_dir,
                filters={'time': {'border': _get_datetime(border),
                                  'op': op}})
            filtered_event_ids = [e.id for e in filtered_events]
            self.assertListEqual(expected_ids, filtered_event_ids)

        time1 = _get_datetime('2030-01-01 01:00')
        time2 = _get_datetime('2030-01-01 02:00')
        time3 = _get_datetime('2030-01-01 03:00')
        sort_key = 'time'
        sort_dir = 'desc'

        db_api.event_create(_get_fake_event_values(id='1', time=time1))
        db_api.event_create(_get_fake_event_values(id='2', time=time2))
        db_api.event_create(_get_fake_event_values(id='3', time=time3))

        check_query('2030-01-01 02:00', 'lt', ['1'])
        check_query('2030-01-01 02:00', 'le', ['2', '1'])
        check_query('2030-01-01 02:00', 'gt', ['3'])
        check_query('2030-01-01 02:00', 'ge', ['3', '2'])
        check_query('2030-01-01 02:00', 'eq', ['2'])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools


class LazyProxy(object):

    def __init__(self, klass, *args, **kwargs):
        self.klass = klass
        self.args = args
        self.kwargs = kwargs
        self.instance = None

    def __getattr__(self, name):
        return functools.partial(self.__run_method, name)

    def __run_method(self, __name, *args, **kwargs):
        if self.instance is None:
            self.instance = self.klass(*self.args, **self.kwargs)
        return getattr(self.instance, __name)(*args, **kwargs)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg

from blazar import context
from blazar.utils.openstack import keystone

CONF = cfg.CONF


def create_trust():
    """Creates trust via Keystone API v3 to use in plugins."""
    trustee_id = keystone.BlazarKeystoneClient().session.get_user_id()

    ctx = context.current()
    user_client = keystone.BlazarKeystoneClient(as_user=True)
    trust = user_client.trusts.create(trustor_user=ctx.user_id,
                                      trustee_user=trustee_id,
                                      impersonation=True,
                                      role_names=ctx.roles,
                                      project=ctx.project_id)
    return trust


def delete_trust(lease):
    """Deletes trust for the specified lease."""
    if lease.trust_id:
        client = keystone.BlazarKeystoneClient(trust_id=lease.trust_id)
        client.trusts.delete(lease.trust_id)


def create_ctx_from_trust(trust_id):
    """Return context built from given trust."""
    ctx = context.current()
    client = keystone.BlazarKeystoneClient(trust_id=trust_id)
    session = client.session

    # use 'with ctx' statement in the place you need context from trust
    return context.BlazarContext(
        user_name=ctx.user_name,
        user_domain_name=ctx.user_domain_name,
        auth_token=session.get_token(),
        project_id=session.get_project_id(),
        service_catalog=(
            ctx.service_catalog or
            session.auth.get_auth_ref(session).service_catalog),
        request_id=ctx.request_id,
        global_request_id=ctx.global_request_id
    )


def use_trust_auth():
    """Decorator creates a keystone trust

    This decorator creates a keystone trust, and adds the trust_id to the
    parameter of the decorated method.
    """
    def decorator(func):

        @functools.wraps(func)
        def wrapped(self, to_update):
            if to_update is not None:
                trust = create_trust()
                if isinstance(to_update, dict):
                    to_update.update({'trust_id': trust.id})
                elif isinstance(to_update, object):
                    setattr(to_update, 'trust_id', trust.id)
            return func(self, to_update)

        return wrapped
    return decorator



# -*- coding: utf-8 -*-
#
# Author: François Rossigneux <francois.rossigneux@inria.fr>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Defines interface for DB access.

Functions in this module are imported into the blazar.db namespace. Call these
functions from blazar.db namespace, not the blazar.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface.

**Related Flags**

:db_backend:  string to lookup in the list of LazyPluggable backends.
              `sqlalchemy` is the only supported backend right now.

:sql_connection:  string specifying the sqlalchemy connection to use, like:
                  `sqlite:///var/lib/blazar/blazar.sqlite`.

"""

from oslo_config import cfg
from oslo_db import api as db_api
from oslo_log import log as logging


_BACKEND_MAPPING = {
    'sqlalchemy': 'blazar.db.sqlalchemy.utils',
}

IMPL = db_api.DBAPI(cfg.CONF.database.backend,
                    backend_mapping=_BACKEND_MAPPING)
LOG = logging.getLogger(__name__)


def setup_db():
    """Set up database, create tables, etc.

    Return True on success, False otherwise
    """
    return IMPL.setup_db()


def drop_db():
    """Drop database.

    Return True on success, False otherwise
    """
    return IMPL.drop_db()


# Helpers for building constraints / equality checks


def constraint(**conditions):
    """Return a constraint object suitable for use with some updates."""
    return IMPL.constraint(**conditions)


def equal_any(*values):
    """Return an equality condition object suitable for use in a constraint.

    Equal_any conditions require that a model object's attribute equal any
    one of the given values.
    """
    return IMPL.equal_any(*values)


def not_equal(*values):
    """Return an inequality condition object suitable for use in a constraint.

    Not_equal conditions require that a model object's attribute differs from
    all of the given values.
    """
    return IMPL.not_equal(*values)


def to_dict(func):
    def decorator(*args, **kwargs):
        res = func(*args, **kwargs)

        if isinstance(res, list):
            return [item.to_dict() for item in res]

        if res:
            return res.to_dict()
        else:
            return None

    return decorator


def get_reservations_by_host_id(host_id, start_date, end_date):
    return IMPL.get_reservations_by_host_id(host_id, start_date, end_date)


def get_reservations_by_host_ids(host_ids, start_date, end_date):
    return IMPL.get_reservations_by_host_ids(host_ids, start_date, end_date)


def get_reservation_allocations_by_host_ids(host_ids, start_date, end_date,
                                            lease_id=None,
                                            reservation_id=None):
    return IMPL.get_reservation_allocations_by_host_ids(host_ids,
                                                        start_date, end_date,
                                                        lease_id,
                                                        reservation_id)


def get_reservation_allocations_by_fip_ids(fip_ids, start_date, end_date,
                                           lease_id=None, reservation_id=None):
    return IMPL.get_reservation_allocations_by_fip_ids(
        fip_ids, start_date, end_date, lease_id, reservation_id)


def get_plugin_reservation(resource_type, resource_id):
    return IMPL.get_plugin_reservation(resource_type, resource_id)


def get_free_periods(resource_id, start_date, end_date, duration,
                     resource_type='host'):
    """Returns a list of free periods."""
    return IMPL.get_free_periods(resource_id, start_date, end_date, duration,
                                 resource_type=resource_type)


def get_reserved_periods(resource_id, start_date, end_date, duration,
                         resource_type='host'):
    """Returns a list of reserved periods."""
    return IMPL.get_reserved_periods(resource_id, start_date, end_date,
                                     duration, resource_type=resource_type)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of SQLAlchemy backend."""

import sys

from oslo_db import exception as common_db_exc
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy.sql.expression import asc
from sqlalchemy.sql.expression import desc

from blazar.db import exceptions as db_exc
from blazar.db.sqlalchemy import facade_wrapper
from blazar.db.sqlalchemy import models

RESOURCE_PROPERTY_MODELS = {
    'physical:host': models.ComputeHostExtraCapability,
}

LOG = logging.getLogger(__name__)


def get_backend():
    """The backend is this module itself."""
    return sys.modules[__name__]


def setup_db():
    try:
        with facade_wrapper.session_for_write(sqlite_fk=True) as session:
            engine = session.get_bind()
            models.Lease.metadata.create_all(engine)
        facade_wrapper._clear_engine()
    except sa.exc.OperationalError as e:
        LOG.error("Database registration exception: %s", e)
        raise
    return True


def drop_db():
    try:
        with facade_wrapper.session_for_write(sqlite_fk=True) as session:
            engine = session.get_bind()
            models.Lease.metadata.drop_all(engine)
        facade_wrapper._clear_engine()
    except Exception as e:
        LOG.error("Database shutdown exception: %s", e)
        raise
    return True


# Helpers for building constraints / equality checks


def constraint(**conditions):
    return Constraint(conditions)


def equal_any(*values):
    return EqualityCondition(values)


def not_equal(*values):
    return InequalityCondition(values)


class Constraint(object):
    def __init__(self, conditions):
        self.conditions = conditions

    def apply(self, model, query):
        for key, condition in self.conditions.items():
            for clause in condition.clauses(getattr(model, key)):
                query = query.filter(clause)
        return query


class EqualityCondition(object):
    def __init__(self, values):
        self.values = values

    def clauses(self, field):
        return sa.or_([field == value for value in self.values])


class InequalityCondition(object):
    def __init__(self, values):
        self.values = values

    def clauses(self, field):
        return [field != value for value in self.values]


# Reservation
def _reservation_get(session, reservation_id):
    query = session.query(models.Reservation)
    return query.filter_by(id=reservation_id).first()


def reservation_get(reservation_id):
    with facade_wrapper.session_for_read() as session:
        return _reservation_get(session, reservation_id)


def reservation_get_all():
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.Reservation)
        return query.all()


def reservation_get_all_by_lease_id(lease_id):
    with facade_wrapper.session_for_read() as session:
        reservations = (session.query(models.Reservation
                                      ).filter_by(lease_id=lease_id))
        return reservations.all()


def reservation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    with facade_wrapper.session_for_read() as session:
        reservation_query = session.query(models.Reservation)
        for name, value in kwargs.items():
            column = getattr(models.Reservation, name, None)
            if column:
                reservation_query = reservation_query.filter(column == value)
        return reservation_query.all()


def reservation_create(values):
    values = values.copy()
    reservation = models.Reservation()
    reservation.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            reservation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=reservation.__class__.__name__, columns=e.columns)

    return reservation_get(reservation.id)


def reservation_update(reservation_id, values):
    with facade_wrapper.session_for_write() as session:
        reservation = _reservation_get(session, reservation_id)
        reservation.update(values)
        reservation.save(session=session)

    return reservation_get(reservation_id)


def reservation_destroy(reservation_id):
    with facade_wrapper.session_for_write() as session:
        reservation = _reservation_get(session, reservation_id)

        if not reservation:
            # raise not found error
            raise db_exc.BlazarDBNotFound(id=reservation_id,
                                          model='Reservation')

        session.delete(reservation)


# Lease
def _lease_get(session, lease_id):
    query = session.query(models.Lease)
    return query.filter_by(id=lease_id).first()


def lease_get(lease_id):
    with facade_wrapper.session_for_read() as session:
        return _lease_get(session, lease_id)


def lease_get_all():
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.Lease)
        return query.all()


def lease_get_all_by_project(project_id):
    raise NotImplementedError


def lease_get_all_by_user(user_id):
    raise NotImplementedError


def lease_list(project_id=None):
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.Lease)
        if project_id is not None:
            query = query.filter_by(project_id=project_id)
        return query.all()


def lease_create(values):
    values = values.copy()
    lease = models.Lease()
    reservations = values.pop("reservations", [])
    events = values.pop("events", [])
    lease.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            lease.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=lease.__class__.__name__, columns=e.columns)

        try:
            for r in reservations:
                reservation = models.Reservation()
                reservation.update({"lease_id": lease.id})
                reservation.update(r)
                reservation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=reservation.__class__.__name__, columns=e.columns)

        try:
            for e in events:
                event = models.Event()
                event.update({"lease_id": lease.id})
                event.update(e)
                event.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=event.__class__.__name__, columns=e.columns)

    return lease_get(lease.id)


def lease_update(lease_id, values):
    with facade_wrapper.session_for_write() as session:
        lease = _lease_get(session, lease_id)
        lease.update(values)
        lease.save(session=session)

    return lease_get(lease_id)


def lease_destroy(lease_id):
    with facade_wrapper.session_for_write() as session:
        lease = _lease_get(session, lease_id)

        if not lease:
            # raise not found error
            raise db_exc.BlazarDBNotFound(id=lease_id, model='Lease')

        session.delete(lease)


# Event
def _event_get(session, event_id):
    query = session.query(models.Event)
    return query.filter_by(id=event_id).first()


def _event_get_all(session):
    query = session.query(models.Event)
    return query


def event_get(event_id):
    with facade_wrapper.session_for_read() as session:
        return _event_get(session, event_id)


def event_get_all():
    with facade_wrapper.session_for_read() as session:
        return _event_get_all(session).all()


def _event_get_sorted_by_filters(session, sort_key, sort_dir, filters):
    """Return an event query filtered and sorted by name of the field."""

    sort_fn = {'desc': desc, 'asc': asc}

    events_query = _event_get_all(session)

    if 'status' in filters:
        events_query = (
            events_query.filter(models.Event.status == filters['status']))
    if 'lease_id' in filters:
        events_query = (
            events_query.filter(models.Event.lease_id == filters['lease_id']))
    if 'event_type' in filters:
        events_query = events_query.filter(models.Event.event_type ==
                                           filters['event_type'])
    if 'time' in filters:
        border = filters['time']['border']
        if filters['time']['op'] == 'lt':
            events_query = events_query.filter(models.Event.time < border)
        elif filters['time']['op'] == 'le':
            events_query = events_query.filter(models.Event.time <= border)
        elif filters['time']['op'] == 'gt':
            events_query = events_query.filter(models.Event.time > border)
        elif filters['time']['op'] == 'ge':
            events_query = events_query.filter(models.Event.time >= border)
        elif filters['time']['op'] == 'eq':
            events_query = events_query.filter(models.Event.time == border)

    events_query = events_query.order_by(
        sort_fn[sort_dir](getattr(models.Event, sort_key))
    )

    return events_query


def event_get_first_sorted_by_filters(sort_key, sort_dir, filters):
    """Return first result for events

    Return the first result for all events matching the filters
    and sorted by name of the field.
    """
    with facade_wrapper.session_for_read() as session:
        return _event_get_sorted_by_filters(
            session, sort_key, sort_dir, filters).first()


def event_get_all_sorted_by_filters(sort_key, sort_dir, filters):
    """Return events filtered and sorted by name of the field."""
    with facade_wrapper.session_for_read() as session:
        return _event_get_sorted_by_filters(
            session, sort_key, sort_dir, filters).all()


def event_create(values):
    values = values.copy()
    event = models.Event()
    event.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            event.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=event.__class__.__name__, columns=e.columns)

    return event_get(event.id)


def event_update(event_id, values):
    with facade_wrapper.session_for_write() as session:
        event = _event_get(session, event_id)
        event.update(values)
        event.save(session=session)

    return event_get(event_id)


def event_destroy(event_id):
    with facade_wrapper.session_for_write() as session:
        event = _event_get(session, event_id)

        if not event:
            # raise not found error
            raise db_exc.BlazarDBNotFound(id=event_id, model='Event')

        session.delete(event)


# ComputeHostReservation
def _host_reservation_get(session, host_reservation_id):
    query = session.query(models.ComputeHostReservation)
    return query.filter_by(id=host_reservation_id).first()


def host_reservation_get(host_reservation_id):
    with facade_wrapper.session_for_read() as session:
        return _host_reservation_get(session, host_reservation_id)


def host_reservation_get_all():
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.ComputeHostReservation)
        return query.all()


def _host_reservation_get_by_reservation_id(session, reservation_id):
    query = session.query(models.ComputeHostReservation)
    return query.filter_by(reservation_id=reservation_id).first()


def host_reservation_get_by_reservation_id(reservation_id):
    with facade_wrapper.session_for_read() as session:
        return _host_reservation_get_by_reservation_id(session, reservation_id)


def host_reservation_create(values):
    values = values.copy()
    host_reservation = models.ComputeHostReservation()
    host_reservation.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host_reservation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host_reservation.__class__.__name__, columns=e.columns)

    return host_reservation_get(host_reservation.id)


def host_reservation_update(host_reservation_id, values):
    with facade_wrapper.session_for_write() as session:
        host_reservation = _host_reservation_get(session, host_reservation_id)
        host_reservation.update(values)
        host_reservation.save(session=session)

    return host_reservation_get(host_reservation_id)


def host_reservation_destroy(host_reservation_id):
    with facade_wrapper.session_for_write() as session:
        host_reservation = _host_reservation_get(session, host_reservation_id)

        if not host_reservation:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=host_reservation_id, model='ComputeHostReservation')

        session.delete(host_reservation)


# InstanceReservation
def instance_reservation_create(values):
    value = values.copy()
    instance_reservation = models.InstanceReservations()
    instance_reservation.update(value)

    with facade_wrapper.session_for_write() as session:
        try:
            instance_reservation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=instance_reservation.__class__.__name__,
                columns=e.columns)

    return instance_reservation_get(instance_reservation.id)


def instance_reservation_get(instance_reservation_id, session=None):
    if not session:
        with facade_wrapper.session_for_read() as session:
            query = session.query(models.InstanceReservations)
            return query.filter_by(id=instance_reservation_id).first()
    query = session.query(models.InstanceReservations)
    return query.filter_by(id=instance_reservation_id).first()


def instance_reservation_update(instance_reservation_id, values):
    with facade_wrapper.session_for_write() as session:
        instance_reservation = instance_reservation_get(
            instance_reservation_id, session)

        if not instance_reservation:
            raise db_exc.BlazarDBNotFound(
                id=instance_reservation_id, model='InstanceReservations')

        instance_reservation.update(values)
        instance_reservation.save(session=session)

    return instance_reservation_get(instance_reservation_id)


def instance_reservation_destroy(instance_reservation_id):
    with facade_wrapper.session_for_write() as session:
        instance = instance_reservation_get(instance_reservation_id)

        if not instance:
            raise db_exc.BlazarDBNotFound(
                id=instance_reservation_id, model='InstanceReservations')
        session.delete(instance)


# ComputeHostAllocation
def _host_allocation_get(session, host_allocation_id):
    query = session.query(models.ComputeHostAllocation)
    return query.filter_by(id=host_allocation_id).first()


def host_allocation_get(host_allocation_id):
    with facade_wrapper.session_for_read() as session:
        return _host_allocation_get(session, host_allocation_id)


def host_allocation_get_all():
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.ComputeHostAllocation)
        return query.all()


def host_allocation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    with facade_wrapper.session_for_read() as session:
        allocation_query = session.query(models.ComputeHostAllocation)
        for name, value in kwargs.items():
            column = getattr(models.ComputeHostAllocation, name, None)
            if column:
                allocation_query = allocation_query.filter(column == value)
        return allocation_query.all()


def host_allocation_create(values):
    values = values.copy()
    host_allocation = models.ComputeHostAllocation()
    host_allocation.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host_allocation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host_allocation.__class__.__name__, columns=e.columns)

    return host_allocation_get(host_allocation.id)


def host_allocation_update(host_allocation_id, values):
    with facade_wrapper.session_for_write() as session:
        host_allocation = _host_allocation_get(session,
                                               host_allocation_id)
        host_allocation.update(values)
        host_allocation.save(session=session)

    return host_allocation_get(host_allocation_id)


def host_allocation_destroy(host_allocation_id):
    with facade_wrapper.session_for_write() as session:
        host_allocation = _host_allocation_get(session,
                                               host_allocation_id)

        if not host_allocation:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=host_allocation_id, model='ComputeHostAllocation')

        session.delete(host_allocation)


# ComputeHost
def _host_get(session, host_id):
    query = session.query(models.ComputeHost)
    return query.filter_by(id=host_id).first()


def _host_get_all(session):
    query = session.query(models.ComputeHost)
    return query


def host_get(host_id):
    with facade_wrapper.session_for_read() as session:
        return _host_get(session, host_id)


def host_list():
    with facade_wrapper.session_for_read() as session:
        return session.query(models.ComputeHost).all()


def host_get_all_by_filters(filters):
    """Returns hosts filtered by name of the field."""

    with facade_wrapper.session_for_read() as session:
        hosts_query = _host_get_all(session)

        if 'status' in filters:
            hosts_query = hosts_query.filter(
                models.ComputeHost.status == filters['status'])

        return hosts_query.all()


def host_get_all_by_queries(queries):
    """Returns hosts filtered by an array of queries.

    :param queries: array of queries "key op value" where op can be
        http://docs.sqlalchemy.org/en/rel_0_7/core/expression_api.html
            #sqlalchemy.sql.operators.ColumnOperators

    """
    with facade_wrapper.session_for_read() as session:
        hosts_query = session.query(models.ComputeHost)

        oper = {
            '<': ['lt', lambda a, b: a >= b],
            '>': ['gt', lambda a, b: a <= b],
            '<=': ['le', lambda a, b: a > b],
            '>=': ['ge', lambda a, b: a < b],
            '==': ['eq', lambda a, b: a != b],
            '!=': ['ne', lambda a, b: a == b],
        }

        hosts = []
        for query in queries:
            try:
                key, op, value = query.split(' ', 2)
            except ValueError:
                raise db_exc.BlazarDBInvalidFilter(query_filter=query)

            column = getattr(models.ComputeHost, key, None)
            if column is not None:
                if op == 'in':
                    filt = column.in_(value.split(','))
                else:
                    if op in oper:
                        op = oper[op][0]
                    try:
                        attr = [e for e in ['%s', '%s_', '__%s__']
                                if hasattr(column, e % op)][0] % op
                    except IndexError:
                        raise db_exc.BlazarDBInvalidFilterOperator(
                            filter_operator=op)

                    if value == 'null':
                        value = None

                    filt = getattr(column, attr)(value)

                hosts_query = hosts_query.filter(filt)
            else:
                # looking for resource properties matches
                extra_filter = (
                    _host_resource_property_query(session)
                    .filter(models.ResourceProperty.property_name == key)
                ).all()

                if not extra_filter:
                    raise db_exc.BlazarDBNotFound(
                        id=key, model='ComputeHostExtraCapability')

                for host, property_name in extra_filter:
                    print(dir(host))
                    if op in oper and oper[op][1](host.capability_value,
                                                  value):
                        hosts.append(host.computehost_id)
                    elif op not in oper:
                        msg = ('Operator %s for resource properties '
                               'not implemented')
                        raise NotImplementedError(msg % op)

                # We must also avoid selecting any host which doesn't have the
                # extra capability present.
                all_hosts = [h.id for h in hosts_query.all()]
                extra_filter_hosts = [h.computehost_id
                                      for h, _ in extra_filter]
                hosts += [h for h in all_hosts if h not in extra_filter_hosts]

        return hosts_query.filter(~models.ComputeHost.id.in_(hosts)).all()


def reservable_host_get_all_by_queries(queries):
    """Returns reservable hosts filtered by an array of queries.

    :param queries: array of queries "key op value" where op can be
        http://docs.sqlalchemy.org/en/rel_0_7/core/expression_api.html
            #sqlalchemy.sql.operators.ColumnOperators

    """
    queries.append('reservable == 1')
    return host_get_all_by_queries(queries)


def unreservable_host_get_all_by_queries(queries):
    """Returns unreservable hosts filtered by an array of queries.

    :param queries: array of queries "key op value" where op can be
        http://docs.sqlalchemy.org/en/rel_0_7/core/expression_api.html
            #sqlalchemy.sql.operators.ColumnOperators

    """

    # TODO(hiro-kobayashi): support the expression 'reservable == False'
    queries.append('reservable == 0')
    return host_get_all_by_queries(queries)


def host_create(values):
    values = values.copy()
    host = models.ComputeHost()
    host.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host.__class__.__name__, columns=e.columns)

    return host_get(host.id)


def host_update(host_id, values):
    with facade_wrapper.session_for_write() as session:
        host = _host_get(session, host_id)
        host.update(values)
        host.save(session=session)

    return host_get(host_id)


def host_destroy(host_id):
    with facade_wrapper.session_for_write() as session:
        host = _host_get(session, host_id)

        if not host:
            # raise not found error
            raise db_exc.BlazarDBNotFound(id=host_id, model='Host')

        session.delete(host)


# ComputeHostExtraCapability

def _host_resource_property_query(session):
    return (
        session.query(models.ComputeHostExtraCapability)
        .join(models.ResourceProperty)
        .add_column(models.ResourceProperty.property_name))


def _host_extra_capability_get(session, host_extra_capability_id):
    query = _host_resource_property_query(session).filter(
        models.ComputeHostExtraCapability.id == host_extra_capability_id)

    return query.first()


def host_extra_capability_get(host_extra_capability_id):
    with facade_wrapper.session_for_read() as session:
        return _host_extra_capability_get(session, host_extra_capability_id)


def _host_extra_capability_get_all_per_host(session, host_id):
    query = _host_resource_property_query(session).filter(
        models.ComputeHostExtraCapability.computehost_id == host_id)

    return query


def host_extra_capability_get_all_per_host(host_id):
    with facade_wrapper.session_for_read() as session:
        return _host_extra_capability_get_all_per_host(session, host_id).all()


def host_extra_capability_create(values):
    values = values.copy()

    resource_property = resource_property_get_or_create(
        'physical:host', values.get('property_name'))

    del values['property_name']
    values['property_id'] = resource_property.id

    host_extra_capability = models.ComputeHostExtraCapability()
    host_extra_capability.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host_extra_capability.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host_extra_capability.__class__.__name__,
                columns=e.columns)

    return host_extra_capability_get(host_extra_capability.id)


def host_extra_capability_update(host_extra_capability_id, values):
    with facade_wrapper.session_for_write() as session:
        host_extra_capability, _ = (
            _host_extra_capability_get(session,
                                       host_extra_capability_id))
        host_extra_capability.update(values)
        host_extra_capability.save(session=session)

    return host_extra_capability_get(host_extra_capability_id)


def host_extra_capability_destroy(host_extra_capability_id):
    with facade_wrapper.session_for_write() as session:
        host_extra_capability = _host_extra_capability_get(
            session, host_extra_capability_id)

        if not host_extra_capability:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=host_extra_capability_id,
                model='ComputeHostExtraCapability')

        session.delete(host_extra_capability[0])


def host_extra_capability_get_all_per_name(host_id, property_name):
    with facade_wrapper.session_for_read() as session:
        query = _host_extra_capability_get_all_per_host(session, host_id)
        return query.filter(
            models.ResourceProperty.property_name == property_name).all()


# ComputeHostResourceInventory

def host_resource_inventory_create(values):
    values = values.copy()

    host_resource_inventory = models.ComputeHostResourceInventory()
    host_resource_inventory.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host_resource_inventory.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host_resource_inventory.__class__.__name__,
                columns=e.columns)

    return None


def host_resource_inventory_get_all_per_host(host_id):
    with facade_wrapper.session_for_read() as session:
        query = session.query(models.ComputeHostResourceInventory)
        return query.filter_by(computehost_id=host_id).all()


# ComputeHostTrait

def host_trait_create(values):
    values = values.copy()

    host_trait = models.ComputeHostTrait()
    host_trait.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            host_trait.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=host_trait.__class__.__name__,
                columns=e.columns)

    return None


# FloatingIP reservation

def fip_reservation_create(fip_reservation_values):
    values = fip_reservation_values.copy()
    fip_reservation = models.FloatingIPReservation()
    fip_reservation.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            fip_reservation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=fip_reservation.__class__.__name__, columns=e.columns)

    return fip_reservation_get(fip_reservation.id)


def _fip_reservation_get(session, fip_reservation_id):
    query = session.query(models.FloatingIPReservation)
    return query.filter_by(id=fip_reservation_id).first()


def fip_reservation_get(fip_reservation_id):
    with facade_wrapper.session_for_read() as session:
        return _fip_reservation_get(session, fip_reservation_id)


def fip_reservation_update(fip_reservation_id, fip_reservation_values):
    with facade_wrapper.session_for_write() as session:
        fip_reservation = _fip_reservation_get(session, fip_reservation_id)
        fip_reservation.update(fip_reservation_values)
        fip_reservation.save(session=session)

    return fip_reservation_get(fip_reservation_id)


def fip_reservation_destroy(fip_reservation_id):
    with facade_wrapper.session_for_write() as session:
        fip_reservation = _fip_reservation_get(session, fip_reservation_id)

        if not fip_reservation:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=fip_reservation_id, model='FloatingIPReservation')

        session.delete(fip_reservation)


# Required FIP

def required_fip_create(required_fip_values):
    values = required_fip_values.copy()
    required_fip = models.RequiredFloatingIP()
    required_fip.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            required_fip.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=required_fip.__class__.__name__, columns=e.columns)

    return required_fip_get(required_fip.id)


def _required_fip_get(session, required_fip_id):
    query = session.query(models.RequiredFloatingIP)
    return query.filter_by(id=required_fip_id).first()


def required_fip_get(required_fip_id):
    with facade_wrapper.session_for_read() as session:
        return _required_fip_get(session, required_fip_id)


def required_fip_update(required_fip_id, required_fip_values):
    with facade_wrapper.session_for_write() as session:
        required_fip = _required_fip_get(session, required_fip_id)
        required_fip.update(required_fip_values)
        required_fip.save(session=session)

    return required_fip_get(required_fip_id)


def required_fip_destroy(required_fip_id):
    with facade_wrapper.session_for_write() as session:
        required_fip = _required_fip_get(session, required_fip_id)

        if not required_fip:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=required_fip_id, model='RequiredFloatingIP')

        session.delete(required_fip)


def required_fip_destroy_by_fip_reservation_id(fip_reservation_id):
    with facade_wrapper.session_for_write() as session:
        required_fips = session.query(
            models.RequiredFloatingIP).filter_by(
            floatingip_reservation_id=fip_reservation_id)
        for required_fip in required_fips:
            required_fip_destroy(required_fip['id'])


# FloatingIP Allocation

def _fip_allocation_get(session, fip_allocation_id):
    query = session.query(models.FloatingIPAllocation)
    return query.filter_by(id=fip_allocation_id).first()


def fip_allocation_get(fip_allocation_id):
    with facade_wrapper.session_for_read() as session:
        return _fip_allocation_get(session, fip_allocation_id)


def fip_allocation_create(allocation_values):
    values = allocation_values.copy()
    fip_allocation = models.FloatingIPAllocation()
    fip_allocation.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            fip_allocation.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=fip_allocation.__class__.__name__, columns=e.columns)

    return fip_allocation_get(fip_allocation.id)


def fip_allocation_get_all_by_values(**kwargs):
    """Returns all entries filtered by col=value."""
    with facade_wrapper.session_for_read() as session:
        allocation_query = session.query(models.FloatingIPAllocation)
        for name, value in kwargs.items():
            column = getattr(models.FloatingIPAllocation, name, None)
            if column:
                allocation_query = allocation_query.filter(column == value)
        return allocation_query.all()


def fip_allocation_destroy(allocation_id):
    with facade_wrapper.session_for_write() as session:
        fip_allocation = _fip_allocation_get(session, allocation_id)

        if not fip_allocation:
            # raise not found error
            raise db_exc.BlazarDBNotFound(
                id=allocation_id, model='FloatingIPAllocation')

        session.delete(fip_allocation)


def fip_allocation_update(allocation_id, allocation_values):
    with facade_wrapper.session_for_write() as session:
        fip_allocation = _fip_allocation_get(session, allocation_id)
        fip_allocation.update(allocation_values)
        fip_allocation.save(session=session)

    return fip_allocation_get(allocation_id)


# Floating IP
def _floatingip_get(session, floatingip_id):
    query = session.query(models.FloatingIP)
    return query.filter_by(id=floatingip_id).first()


def _floatingip_get_all(session):
    query = session.query(models.FloatingIP)
    return query


def fip_get_all_by_queries(queries):
    """Returns Floating IPs filtered by an array of queries.

    :param queries: array of queries "key op value" where op can be
        http://docs.sqlalchemy.org/en/rel_0_7/core/expression_api.html
            #sqlalchemy.sql.operators.ColumnOperators

    """
    with facade_wrapper.session_for_read() as session:
        fips_query = session.query(models.FloatingIP)

    oper = {
        '<': ['lt', lambda a, b: a >= b],
        '>': ['gt', lambda a, b: a <= b],
        '<=': ['le', lambda a, b: a > b],
        '>=': ['ge', lambda a, b: a < b],
        '==': ['eq', lambda a, b: a != b],
        '!=': ['ne', lambda a, b: a == b],
    }

    for query in queries:
        try:
            key, op, value = query.split(' ', 2)
        except ValueError:
            raise db_exc.BlazarDBInvalidFilter(query_filter=query)

        column = getattr(models.FloatingIP, key, None)
        if column is not None:
            if op == 'in':
                filt = column.in_(value.split(','))
            else:
                if op in oper:
                    op = oper[op][0]
                try:
                    attr = [e for e in ['%s', '%s_', '__%s__']
                            if hasattr(column, e % op)][0] % op
                except IndexError:
                    raise db_exc.BlazarDBInvalidFilterOperator(
                        filter_operator=op)

                if value == 'null':
                    value = None

                filt = getattr(column, attr)(value)

            fips_query = fips_query.filter(filt)
        else:
            raise db_exc.BlazarDBInvalidFilter(query_filter=query)

    return fips_query.all()


def reservable_fip_get_all_by_queries(queries):
    """Returns reservable fips filtered by an array of queries.

    :param queries: array of queries "key op value" where op can be
        http://docs.sqlalchemy.org/en/rel_0_7/core/expression_api.html
            #sqlalchemy.sql.operators.ColumnOperators

    """
    queries.append('reservable == 1')
    return fip_get_all_by_queries(queries)


def floatingip_get(floatingip_id):
    with facade_wrapper.session_for_read() as session:
        return _floatingip_get(session, floatingip_id)


def floatingip_list():
    with facade_wrapper.session_for_read() as session:
        return session.query(models.FloatingIP).all()


def floatingip_create(values):
    values = values.copy()
    floatingip = models.FloatingIP()
    floatingip.update(values)

    with facade_wrapper.session_for_write() as session:
        try:
            floatingip.save(session=session)
        except common_db_exc.DBDuplicateEntry as e:
            # raise exception about duplicated columns (e.columns)
            raise db_exc.BlazarDBDuplicateEntry(
                model=floatingip.__class__.__name__, columns=e.columns)

    return floatingip_get(floatingip.id)


def floatingip_destroy(floatingip_id):
    with facade_wrapper.session_for_write() as session:
        floatingip = _floatingip_get(session, floatingip_id)

        if not floatingip:
            # raise not found error
            raise db_exc.BlazarDBNotFound(id=floatingip_id, model='FloatingIP')

        session.delete(floatingip)


# Resource Properties

def _resource_property_get(session, resource_type, property_name):
    query = (
        session.query(models.ResourceProperty)
        .filter_by(resource_type=resource_type)
        .filter_by(property_name=property_name))

    return query.first()


def resource_property_get(resource_type, property_name):
    with facade_wrapper.session_for_read() as session:
        return _resource_property_get(session, resource_type, property_name)


def resource_properties_list(resource_type):
    if resource_type not in RESOURCE_PROPERTY_MODELS:
        raise db_exc.BlazarDBResourcePropertiesNotEnabled(
            resource_type=resource_type)

    with facade_wrapper.session_for_read() as session:
        resource_model = RESOURCE_PROPERTY_MODELS[resource_type]
        query = session.query(
            models.ResourceProperty.property_name,
            models.ResourceProperty.private,
            resource_model.capability_value).join(resource_model).distinct()

        return query.all()


def _resource_property_create(session, values):
    values = values.copy()

    resource_property = models.ResourceProperty()
    resource_property.update(values)

    try:
        resource_property.save(session=session)
    except common_db_exc.DBDuplicateEntry as e:
        # raise exception about duplicated columns (e.columns)
        raise db_exc.BlazarDBDuplicateEntry(
            model=resource_property.__class__.__name__,
            columns=e.columns)

    return resource_property_get(values.get('resource_type'),
                                 values.get('property_name'))


def resource_property_create(values):
    with facade_wrapper.session_for_write() as session:
        return _resource_property_create(session, values)


def resource_property_update(resource_type, property_name, values):
    if resource_type not in RESOURCE_PROPERTY_MODELS:
        raise db_exc.BlazarDBResourcePropertiesNotEnabled(
            resource_type=resource_type)

    values = values.copy()
    with facade_wrapper.session_for_write() as session:
        resource_property = _resource_property_get(
            session, resource_type, property_name)

        if not resource_property:
            raise db_exc.BlazarDBInvalidResourceProperty(
                property_name=property_name,
                resource_type=resource_type)

        resource_property.update(values)
        resource_property.save(session=session)

    return resource_property_get(resource_type, property_name)


def _resource_property_get_or_create(session, resource_type, property_name):
    resource_property = _resource_property_get(
        session, resource_type, property_name)

    if resource_property:
        return resource_property
    else:
        rp_values = {
            'resource_type': resource_type,
            'property_name': property_name}

        return resource_property_create(rp_values)


def resource_property_get_or_create(resource_type, property_name):
    with facade_wrapper.session_for_write() as session:
        return _resource_property_get_or_create(
            session, resource_type, property_name)



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import threading

from oslo_db.sqlalchemy import enginefacade

_CONTEXT = threading.local()

_engine_facade = None


def session_for_read():
    return _get_facade().reader.using(_CONTEXT)


def session_for_write(sqlite_fk=False):
    return _get_facade(sqlite_fk=sqlite_fk).writer.using(_CONTEXT)


def _clear_engine():
    global _engine_facade
    _engine_facade = None


def _get_facade(sqlite_fk=False):
    global _engine_facade
    if not _engine_facade:
        ctx = enginefacade.transaction_context()
        ctx.configure(sqlite_fk=sqlite_fk)
        _engine_facade = ctx

    return _engine_facade



# Copyright 2022 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""resource property

Revision ID: 02e2f2186d98
Revises: f4084140f608
Create Date: 2020-04-17 15:51:40.542459

"""

# revision identifiers, used by Alembic.
revision = '02e2f2186d98'
down_revision = 'f4084140f608'

import uuid

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql


def upgrade():
    op.create_table('resource_properties',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('resource_type', sa.String(255), nullable=False),
                    sa.Column('property_name', sa.String(255),
                              nullable=False),
                    sa.Column('private', sa.Boolean, nullable=False,
                              server_default=sa.false()),
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('resource_type', 'property_name'))

    if op.get_bind().engine.name != 'sqlite':
        connection = op.get_bind()

        host_query = connection.execute(sa.text("""
            SELECT DISTINCT "physical:host", capability_name
            FROM computehost_extra_capabilities;"""))

        capability_values = [
            (str(uuid.uuid4()), resource_type, capability_name)
            for resource_type, capability_name
            in host_query.fetchall()]

        if capability_values:
            insert = """
                INSERT INTO resource_properties
                (id, resource_type, property_name)
                VALUES {};"""
            connection.execute(sa.text(
                insert.format(', '.join(map(str, capability_values)))))

        op.add_column('computehost_extra_capabilities',
                      sa.Column('property_id', sa.String(length=255),
                                nullable=False))

        connection.execute(sa.text("""
            UPDATE computehost_extra_capabilities c
            LEFT JOIN resource_properties e
            ON e.property_name = c.capability_name
            SET c.property_id = e.id;"""))

        op.create_foreign_key('computehost_resource_property_id_fk',
                              'computehost_extra_capabilities',
                              'resource_properties', ['property_id'], ['id'])
        op.drop_column('computehost_extra_capabilities', 'capability_name')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('computehost_extra_capabilities',
                  sa.Column('capability_name', mysql.VARCHAR(length=64),
                            nullable=False))

    if op.get_bind().engine.name != 'sqlite':
        connection = op.get_bind()
        connection.execute(sa.text("""
            UPDATE computehost_extra_capabilities c
            LEFT JOIN resource_properties e
            ON e.id=c.property_id
            SET c.capability_name = e.property_name;"""))
        op.drop_constraint('computehost_resource_property_id_fk',
                           'computehost_extra_capabilities',
                           type_='foreignkey')
        op.drop_column('computehost_extra_capabilities', 'property_id')
        op.drop_table('resource_properties')
